/////////////////////////////////////////////////////////////////////////////
// Name:        MyFrame.cpp
// Purpose:    App, Frame and Layout stuff
// Part of:       4Pane
// Author:       David Hart
// Copyright:   (c) 2007 David Hart
// Licence:     GPL v3
/////////////////////////////////////////////////////////////////////////////

// For compilers that support precompilation, includes "wx/wx.h". 
#include "wx/wxprec.h"

#ifndef WX_PRECOMP
    #include "wx/log.h"

    #include "wx/app.h"
    #include "wx/frame.h"

    #include "wx/scrolwin.h"
    #include "wx/menu.h"
	#include <wx/dirctrl.h>

	#include "wx/cmdline.h"
	#include "wx/dragimag.h"
	#include "wx/config.h"
#endif
	
#if wxVERSION_NUMBER > 2503	
	#include <wx/stdpaths.h>
#endif

#if defined __WXGTK__
	#include <gtk/gtk.h>
#endif
	
#include "wx/intl.h"

#include "Externs.h"
#include "Tools.h"
#include "MyTreeCtrl.h"
#include "MyDirs.h"
#include "MyFiles.h"
#include "MyGenericDirCtrl.h"
#include "Archive.h"
#include "Accelerators.h"
#include "MyFrame.h"
#include "Filetypes.h"
#include "config.h"
#include "Misc.h"
#include "Configure.h"
#include "Redo.h"
#include "Devices.h"
#include "Bookmarks.h"

#include "bitmaps/include/4PaneIcon32.xpm"
#include "bitmaps/include/cut.xpm"
#include "bitmaps/include/copy.xpm"
#include "bitmaps/include/paste.xpm"
#include "bitmaps/include/undo.xpm"
#include "bitmaps/include/redo.xpm"
#include "bitmaps/include/NewTab.xpm"
#include "bitmaps/include/DelTab.xpm"
   
 
// ----------------------------------------------------------------------------
// MyApp 
// ----------------------------------------------------------------------------
wxString myNullString(wxT(""));											// For use when passing wxString& params and you want a null default

wxWindowID NextID = wxID_HIGHEST+5000;					// An int to generate "unique" window IDs.  Inc at each use.  Starts +5k to leave room for xrc ids

enum helpcontext HelpContext = HCunknown;					// What Help to show on F1

IMPLEMENT_APP(MyApp)

bool MyApp::OnInit()
{
m_locale.Init(); 																	// Initialize the catalog we'll be using
wxLocale::AddCatalogLookupPathPrefix(wxT("./locale"));	// Just in case we weren't 'make install'ed
wxLocale::AddCatalogLookupPathPrefix(wxT("/usr/share/locale"));	// Just in case we were installed into /usr/local/

if ( ! m_locale.AddCatalog(wxT("4Pane")) )
	m_locale.AddCatalog(wxT("4pane"));								// Hedge bets re our spelling

if ( ! ParseCmdline() )  return false;

if ( ! Configure::FindIni() ) return false;							// Make sure there's a config file, resources etc. If not, show wizard and create one, find them etc, or abort

wxImage::AddHandler(new wxXPMHandler);						// Load handler for the icons 
wxImage::AddHandler(new wxPNGHandler);						//   & for the copier images

wxXmlResource::Get()->InitAllHandlers();
#ifdef __WXGTK20__
	wxXmlResource::Get()->Load(RCDIR + wxT("gtk2.xrc"));						// These contain things that differ between 1.2 and 2
#else
	wxXmlResource::Get()->Load(RCDIR + wxT("gtk1.2.xrc"));
#endif

wxXmlResource::Get()->Load(RCDIR + wxT("dialogs.xrc"));						//  & dialogues
wxXmlResource::Get()->Load(RCDIR + wxT("moredialogs.xrc"));				//  & more dialogues (eg Archive & Tools ones)
wxXmlResource::Get()->Load(RCDIR + wxT("configuredialogs.xrc"));		//  Guess what

long height = wxConfigBase::Get()->Read( wxT("/Misc/AppHeight"), 400l );
long width = wxConfigBase::Get()->Read( wxT("/Misc/AppWidth"), 600l );

frame = new MyFrame( wxSize(width, height) );
wxString title = GetAppName();											// In case this isn't the normal one
if ( getuid()==0 ) title += wxT(" (superuser)");		// It's nice to know if we're root
frame->SetTitle( title );
frame->GetMenuBar()->Check( SHCUT_SHOW_RECURSIVE_SIZE, SHOW_RECURSIVE_FILEVIEW_SIZE );
frame->GetMenuBar()->Check( SHCUT_RETAIN_REL_TARGET, RETAIN_REL_TARGET );
return true;
}

#if wxVERSION_NUMBER < 2900	
static const wxCmdLineEntryDesc cmdLineDesc[] =
{       
	{ wxCMD_LINE_PARAM,  NULL, NULL, wxT("Display from this filepath (optional)"), wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL | wxCMD_LINE_PARAM_MULTIPLE },
	{ wxCMD_LINE_OPTION, wxT("d"), wxT("home-dir"),  wxT("Use this as the home directory (optional)") },
	{ wxCMD_LINE_OPTION, wxT("c"), wxT("config-fpath"),   wxT("Use this filepath as 4Pane's configuration file (optional; if the filepath doesn't exist, it will be created)") },
	{ wxCMD_LINE_SWITCH, wxT("h"), wxT("help"), wxT("Show this help message"), wxCMD_LINE_VAL_NONE, wxCMD_LINE_OPTION_HELP },

	{ wxCMD_LINE_NONE }
};
#else
static const wxCmdLineEntryDesc cmdLineDesc[] =
{       
	{ wxCMD_LINE_PARAM,  NULL, NULL, "Display from this filepath (optional)", wxCMD_LINE_VAL_STRING, wxCMD_LINE_PARAM_OPTIONAL | wxCMD_LINE_PARAM_MULTIPLE },
	{ wxCMD_LINE_OPTION, "d", "home-dir",  "Use this as the home directory (optional)" },
	{ wxCMD_LINE_OPTION, "c", "config-fpath",   "Use this filepath as 4Pane's configuration file (optional; if the filepath doesn't exist, it will be created)" },
	{ wxCMD_LINE_SWITCH, "h", "help", "Show this help message", wxCMD_LINE_VAL_NONE, wxCMD_LINE_OPTION_HELP },

	{ wxCMD_LINE_NONE }
};
#endif

bool MyApp::ParseCmdline()
{
wxCmdLineParser parser(cmdLineDesc, argc, argv);
switch ( parser.Parse() )
	{	case -1:   return false;		// The parameter -h was passed, help was given, so abort the app
        case 0:	break;					// OK, so break to deal with any parameters etc
        default:  return false;		// Some syntax error occurred. Abort
	}

					// If the app was launched via the 4pane->4Pane symlink, change the appname; otherwise resources aren't located
wxString appname( argv[0] ); appname = appname.AfterLast(wxFILE_SEP_PATH);
if ( appname==wxT("4pane") ) SetAppName( wxT("4Pane") );

if ( parser.GetParamCount() > 0 )		startdir0 = parser.GetParam(0);		// We can cope with up to 2 startdirs
if ( parser.GetParamCount() > 1 )		startdir1 = parser.GetParam(1);
		
if ( parser.Found( wxT("d"), &HOME ) ) 								// See if there's an alternative home directory (useful perhaps if we're on a LiveCD)
	{	if ( ! wxDirExists( HOME ) )											// It would normally be a dir, but check
		 	{	HOME = HOME.BeforeLast(wxFILE_SEP_PATH);	// See if it was /foo/bar/.Not4Pane or something
				if ( ! wxDirExists( HOME ) )	HOME = wxGetHomeDir();	// If the parameter is rubbish, this effectively ignores it
			}
	}
 else	HOME = wxGetHomeDir();										// The obvious default

if ( parser.Found( wxT("c"), &ConfigFilepath ) ) 				// See if there's an alternative config file, or a desire for one (again, LiveCD)
	{	if ( wxDirExists( ConfigFilepath ) )	ConfigFilepath << wxT("/.4Pane");	// It would normally be a file, but check and, if need be, add the name
		if ( ! wxIsAbsolutePath(ConfigFilepath) ) 					// Check that we weren't passed a filename, rather than a filepath
			{	wxString hm = HOME;
				if ( hm.Right(1) != wxFILE_SEP_PATH ) hm << wxFILE_SEP_PATH;
				ConfigFilepath = hm + ConfigFilepath;
			}
		wxString path = ConfigFilepath.BeforeLast( wxFILE_SEP_PATH ); FileData temp( path );	// Now check for permission/ro filesystem
		if ( ! temp.CanTHISUserWrite() )
			{	wxLogError(_("Sorry, you don't have permission to write to the directory containing your chosen configuration-file; aborting.")); return false; }
	}

#if wxVERSION_NUMBER < 2600		// Before this, the only way to make user-defined menu shortcut alterations 'take' is to restart
AbsoluteFilepath = argv[0];												// The relative path, on my system at least
if ( ! wxIsAbsolutePath(AbsoluteFilepath) )							// This might then be 4Pane (in which case we don't need to do anything else
	if ( AbsoluteFilepath.GetChar(0) == wxT('.') )					// or ./4pane; in which case prepend the cwd
		{	AbsoluteFilepath = AbsoluteFilepath.Mid(1);
			wxString cwd = wxGetCwd();
			if (cwd.Right(1) != wxFILE_SEP_PATH) cwd << wxFILE_SEP_PATH;
			AbsoluteFilepath = cwd + AbsoluteFilepath;
		}
#endif

#if wxVERSION_NUMBER > 2503	
	StdDataDir = wxStandardPaths::Get().GetDataDir();		// This will probably be /usr/local/share/4Pane
	size_t pre = StdDataDir.find( wxT("4Pane") );
	if ( pre != wxString::npos )	StdDocsDir = StdDataDir.Left(pre) + wxT("doc/") + wxT("4Pane");
	  else StdDocsDir = StdDataDir + wxT("/doc/");				// Shouldn't happen, but this is a reasonable default
#else
	StdDataDir = wxT("/usr/local/share/4Pane");
	StdDocsDir = wxT("/usr/local/share/doc/4Pane");
#endif

return true;
}

int MyApp::FilterEvent(wxEvent& event)
{
if ( event.GetEventType()==wxEVT_KEY_DOWN && ((wxKeyEvent&)event).GetKeyCode()==WXK_F1 && frame!=NULL )
	{ frame->OnHelpF1( (wxKeyEvent&)event ); return true; }

return -1;
}

#if wxVERSION_NUMBER < 2600
void MyApp::RestartApplic()	// Used after shortcuts have been changed in wxWidgets < 2.6.0
{
Yield(true);																			// Just in case anything's happening
execv( AbsoluteFilepath.mb_str(wxConvUTF8), (char* const*)argv );		// Restart 4Pane
}
#endif

#if defined(__WXX11__)	
	#include "wx/x11/private.h"
bool MyApp::ProcessXEvent(WXEvent* _event)	// Grab selection XEvents
{
if ( wxApp::ProcessXEvent(_event) ) return true;				// First let the standard method happen
 
XEvent* event = (XEvent*) _event;										// The next bit is duplicated from wxApp::ProcessXEvent
wxWindow* win = NULL;
Window window = XEventGetWindow(event);
win = wxGetWindowFromTable(window);
if (!win)
	{
#if wxUSE_TWO_WINDOWS
		win = wxGetClientWindowFromTable(window);
		if (!win)
#endif
		return false;
	}
#if wxVERSION_NUMBER < 2900		// Makes the mousewheel work
if ( event->type == ButtonPress )
	if ( event->xbutton.button == Button4 || event->xbutton.button == Button5 )
		{	wxMouseEvent wxevent;
			wxevent.SetTimestamp(XButtonEventGetTime(event));
			wxevent.SetEventType( wxEVT_MOUSEWHEEL );
			wxevent.m_x = XButtonEventGetX(event);
			wxevent.m_y = XButtonEventGetY(event);
			wxevent.m_leftDown =  XButtonEventLIsDown(event);
			wxevent.m_middleDown = XButtonEventMIsDown(event);
			wxevent.m_rightDown = XButtonEventRIsDown(event);
			wxevent.m_shiftDown = XButtonEventShiftIsDown(event);
			wxevent.m_controlDown = XButtonEventCtrlIsDown(event);
			wxevent.m_altDown = XButtonEventAltIsDown((event));
			wxevent.m_metaDown = XButtonEventMetaIsDown(event);
			wxevent.m_linesPerAction = LINES_PER_MOUSEWHEEL_SCROLL;
			wxevent.m_wheelDelta = 120;
			wxevent.m_wheelRotation = (event->xbutton.button == Button4 ? 120 : -120);

			wxevent.SetId(win->GetId());
			wxevent.SetEventObject(win);

			return win->GetEventHandler()->ProcessEvent( wxevent );
		}
#endif //wxVERSION_NUMBER < 2900

if (  win->GetName() == wxT("TerminalEm")							// Kludge, but I can't get RTTI to work here. Too many levels of subclassing?
										|| ( frame && win == frame->m_tbText ) )
	{ ((TextCtrlBase*)win)->OnXevent( *event); return true; }	// This looks for Selection events, to implement PRIMARY and CLIPBOARD for  textctrls

return false;
}
#endif

int MyApp::OnExit()
{
//wxLogTrace("Starting OnExit()");
delete wxConfigBase::Set((wxConfigBase *) NULL);
return TRUE;
}


// ----------------------------------------------------------------------------
// MyFrame 
// ----------------------------------------------------------------------------

BEGIN_EVENT_TABLE(MyFrame, wxFrame)
	EVT_MENU(SHCUT_UNDO, MyFrame::OnUndo)
	EVT_BUTTON(XRCID("IDM_TOOLBAR_LargeDropdown1"), MyFrame::OnUndoSidebar)
	EVT_MENU(SHCUT_REDO, MyFrame::OnRedo)
	EVT_BUTTON(XRCID("IDM_TOOLBAR_LargeDropdown2"), MyFrame::OnRedoSidebar)
	EVT_MENU(SHCUT_NEWTAB, MyFrame::OnAppendTab)
	EVT_MENU(SHCUT_INSERTTAB, MyFrame::OnInsTab)
	EVT_MENU(SHCUT_DELTAB, MyFrame::OnDelTab)
	EVT_MENU(SHCUT_RENAMETAB, MyFrame::OnRenTab)
	EVT_MENU(SHCUT_DUPLICATETAB, MyFrame::OnDuplicateTab)
#if defined __WXGTK__
	EVT_MENU(SHCUT_ALWAYS_SHOW_TAB, MyFrame::OnAlwaysShowTab)
	EVT_MENU(SHCUT_SAME_TEXTSIZE_TAB, MyFrame::OnSameTabSize)
#endif
	EVT_MENU(SHCUT_REPLICATE, MyFrame::OnReplicate)
	EVT_MENU(SHCUT_SWAPPANES, MyFrame::OnSwapPanes)
	
	EVT_MENU(SHCUT_CUT, MyFrame::OnCut)
	EVT_MENU(SHCUT_COPY, MyFrame::OnCopy)
	EVT_MENU(SHCUT_PASTE, MyFrame::OnPaste)
	EVT_MENU(SHCUT_NEW, MyFrame::OnNew)
	EVT_MENU(SHCUT_HARDLINK, MyFrame::OnHardLink)
	EVT_MENU(SHCUT_SOFTLINK, MyFrame::OnSoftLink)
	EVT_MENU(SHCUT_TRASH, MyFrame::OnTrash)
	EVT_MENU(SHCUT_DELETE, MyFrame::OnDelete)	
	EVT_MENU(SHCUT_RENAME, MyFrame::OnRename)
	EVT_MENU(SHCUT_DUP, MyFrame::OnDup)
	EVT_MENU(SHCUT_PROPERTIES, MyFrame::OnProperties)	

	EVT_MENU(SHCUT_SPLITPANE_VERTICAL, MyFrame::SplitVertical)	
	EVT_MENU(SHCUT_SPLITPANE_HORIZONTAL, MyFrame::SplitHorizontal)	
	EVT_MENU(SHCUT_SPLITPANE_UNSPLIT, MyFrame::Unsplit)	
	EVT_MENU_RANGE(SHCUT_TABTEMPLATERANGE_START, SHCUT_TABTEMPLATERANGE_FINISH, MyFrame::OnTabTemplateLoadMenu)
	EVT_MENU(SHCUT_TAB_SAVEAS_TEMPLATE, MyFrame::OnSaveAsTemplate)
	EVT_MENU(SHCUT_TAB_DELETE_TEMPLATE, MyFrame::OnDeleteTemplate)
	
	EVT_MENU(SHCUT_LAUNCH_TERMINAL, MyFrame::OnLaunchTerminal)
	
	EVT_MENU(SHCUT_ADD_TO_BOOKMARKS, MyFrame::OnAddToBookmarks)
	EVT_MENU(SHCUT_MANAGE_BOOKMARKS, MyFrame::OnManageBookmarks)
	EVT_MENU_RANGE(ID_FIRST_BOOKMARK, ID__LAST_BOOKMARK, MyFrame::OnBookmark)
	
	EVT_MENU(SHCUT_ARCHIVE_EXTRACT, MyFrame::OnArchiveExtract)
	EVT_MENU(SHCUT_ARCHIVE_CREATE, MyFrame::OnArchiveCreate)
	EVT_MENU(SHCUT_ARCHIVE_APPEND, MyFrame::OnArchiveAppend)	
	EVT_MENU(SHCUT_ARCHIVE_TEST, MyFrame::OnArchiveTest)	
	EVT_MENU(SHCUT_ARCHIVE_COMPRESS, MyFrame::OnArchiveCompress)

	EVT_MENU(SHCUT_MOUNT_MOUNTPARTITION, MyFrame::OnMountPartition)
	EVT_MENU(SHCUT_MOUNT_UNMOUNTPARTITION, MyFrame::OnUnMountPartition)
	EVT_MENU(SHCUT_MOUNT_MOUNTISO, MyFrame::OnMountIso)
	EVT_MENU(SHCUT_MOUNT_MOUNTNFS, MyFrame::OnMountNfs)
	EVT_MENU(SHCUT_MOUNT_MOUNTSAMBA, MyFrame::OnMountSamba)
	EVT_MENU(SHCUT_MOUNT_UNMOUNTNETWORK, MyFrame::OnUnMountNetwork)
	EVT_MENU(SHCUT_COMMANDLINE, MyFrame::OnCommandLine)
	EVT_MENU(SHCUT_TERMINAL_EMULATOR, MyFrame::OnTermEm)
	EVT_MENU(SHCUT_TOOL_LOCATE, MyFrame::OnLocate)
	EVT_MENU(SHCUT_TOOL_FIND, MyFrame::OnFind)
	EVT_MENU(SHCUT_TOOL_GREP, MyFrame::OnGrep)
	
	EVT_MENU_RANGE(SHCUT_TOOLS_LAUNCH,SHCUT_TOOLS_LAUNCH_LAST, MyFrame::OnToolsLaunch)
	EVT_MENU(SHCUT_TOOLS_REPEAT, MyFrame::OnToolsRepeat)
	
	EVT_MENU(SHCUT_EMPTYTRASH, MyFrame::OnEmptyTrash)
	EVT_MENU(SHCUT_EMPTYDELETED, MyFrame::OnEmptyTrash)
	
	EVT_MENU(SHCUT_SHOW_RECURSIVE_SIZE, MyFrame::ToggleFileviewSizetype)
	EVT_MENU(SHCUT_RETAIN_REL_TARGET, MyFrame::ToggleRetainRelSymlinkTarget)
	EVT_MENU(SHCUT_CONFIGURE, MyFrame::OnConfigure)
	
	EVT_MENU(SHCUT_SAVETABS, MyFrame::OnSaveTabs)
	
	EVT_MENU(SHCUT_HELP, MyFrame::OnHelpContents)
	EVT_MENU(SHCUT_FAQ, MyFrame::OnHelpFAQs)
	EVT_MENU(SHCUT_ABOUT, MyFrame::OnHelpAbout)
	
	EVT_MENU(SHCUT_CONFIG_SHORTCUTS, MyFrame::OnConfigureShortcuts)
	
	EVT_TEXT_ENTER(ID_FILEPATH, MyFrame::OnTextFilepathEnter)
	
	EVT_MOTION(MyFrame::DoMutex)		
	EVT_LEFT_UP(MyFrame::DoMutex)
	EVT_RIGHT_DOWN(MyFrame::OnDnDRightclick)
	EVT_MOUSEWHEEL(MyFrame::OnMouseWheel)

#if wxVERSION_NUMBER >= 2800
	EVT_MOUSE_CAPTURE_LOST(MyFrame::OnMouseCaptureLost)
#endif
    EVT_MENU(SPLIT_VERTICAL, MyFrame::SplitVertical)
    EVT_MENU(SPLIT_HORIZONTAL, MyFrame::SplitHorizontal)
    EVT_MENU(SPLIT_UNSPLIT, MyFrame::Unsplit)
		
    EVT_MENU(SHCUT_EXIT, MyFrame::OnQuit)
	
	EVT_MENU(SHCUT_FILTER, MyFrame::OnFilter)
	EVT_MENU(SHCUT_TOGGLEHIDDEN, MyFrame::OnToggleHidden)
	EVT_RELOADSHORTCUTS(MyFrame::ReconfigureShortcuts)
	
	EVT_UPDATE_UI_RANGE(SHCUT_PASTE,SHCUT_NEW, MyFrame::DoMiscUI)
	EVT_UPDATE_UI_RANGE(SHCUT_CUT,SHCUT_OPENWITH_KDESU, MyFrame::DoQueryEmptyUI)	// Context menu items that depend on a selected item
	EVT_UPDATE_UI_RANGE(SHCUT_ALWAYS_SHOW_TAB,	SHCUT_SAME_TEXTSIZE_TAB, MyFrame::DoMiscTabUI)
	EVT_UPDATE_UI_RANGE(SHCUT_SPLITPANE_VERTICAL, SHCUT_SPLITPANE_UNSPLIT, MyFrame::DoSplitPaneUI)
	EVT_UPDATE_UI( SHCUT_TAB_DELETE_TEMPLATE, MyFrame::DoTabTemplateUI)
	EVT_UPDATE_UI( SHCUT_TOOLS_REPEAT, MyFrame::DoRepeatCommandUI)
	EVT_UPDATE_UI_RANGE(SHCUT_SHOW_COL_EXT, SHCUT_SHOW_COL_LINK, MyFrame::DoColViewUI)
END_EVENT_TABLE()

wxString CreateSubgroupName( size_t no, size_t total )	// Creates a unique subgroup name in ascending alphabetical order
{																							// So if total is between 26^2 & 26^3, names will start aaa, then aab, aac etc
wxString name;
if ( !total ) return name;														// GIGO

size_t magnitude = 1;															// Work out how many letters we'll need for the whole subgroup ie 1st name may be a, aa, aaa etc
size_t  tot = total;
while ( tot /= 26 ) 		magnitude *= 26;

for ( size_t divisor = magnitude;  divisor > 0; divisor /= 26 )	// Now work out which each of these letters should be for this no
	{	name << wxChar(wxT('a') + (no / divisor));
		no %= divisor;
	}

return name;
}

MyFrame::MyFrame(wxSize size)  : wxFrame(NULL, -1, wxT("4Pane"),  wxDefaultPosition, size,
																														 wxDEFAULT_FRAME_STYLE | wxNO_FULL_REPAINT_ON_RESIZE, wxT("MyFrame"))
{
SelectionAvailable = false;

m_tbText = NULL;																	// In case this isn't going to be used, start with it null for ease of testing

configure =  new class Configure;
configure->LoadConfiguration();

int widths[ 4 ];
ConfigureMisc::LoadStatusbarWidths( widths );		// Load (or set from defaults) the widths of the 4 statusbar sections.  Proportional and, yes, the numbers are *supposed* to be negative
SetStatusBar( CreateStatusBar(4) );
SetStatusWidths(4, widths);

AccelList = new class AcceleratorList( this ); AccelList->Init();
CreateAcceleratorTable();

mainframe = this;																	// Set up the static ptr to the frame
 
SetIcon( wxICON( FourPaneIcon32 ) );

if ( ! HELPDIR.IsEmpty() )														// If we couldn't find a valid HELPDIR, limp along helplessly
	{	Help = new wxHtmlHelpController( wxHF_TOOLBAR | wxHF_CONTENTS | wxHF_MERGE_BOOKS );
		if ( Help->Initialize(HELPDIR) )
				{ wxLogError(_("Cannot initialize the help system; aborting.")); return; }

		Help->AddBook(wxFileName(HELPDIR + wxT("/Chapt.hhp")));		
	}

wxMenuBar* MenuBar = new wxMenuBar();							// Now we can create the bars
CreateMenuBar( MenuBar ); SetMenuBar( MenuBar );

toolbar = new wxToolBar(this, ID_FRAMETOOLBAR ,wxPoint(0,0), wxSize(-1,MAINTOOLBARWIDTH));		// Toolbar pt1
LoadToolbarButtons();

panelette = new wxPanel(this, -1, wxDefaultPosition, wxSize(-1,MAINTOOLBARWIDTH), wxTAB_TRAVERSAL, wxT("TBPanel"));	// & pt2, for MyBitmapButtons
sizerTB = new wxBoxSizer( wxHORIZONTAL );
sizerTB->Add( toolbar, 0, wxEXPAND );
sizerTB->Add( panelette, 1, wxEXPAND );

sizerMain = new wxBoxSizer( wxVERTICAL );
sizerMain->Add( sizerTB, 0, wxEXPAND );

AddControls();
Layout = new LayoutWindows(this);										// This does the rest of the layout things
Layout->Setup();

#if defined(__WXGTK20__)
					// gtk2 steals keydown events for any shortcuts!  So if you use DEL as ->Trash, textctrls never get DEL keypresses.  Ditto Copy/Paste.  This is a workaround (see Tools.cpp). 
PushEventHandler(new EventDistributor);
#endif
						// The next line came from a refused bugreport [ 1181018 ] by Paul Koning.  Without it, the F10 key always opens the File menu!
#if defined(__WXGTK20__)
wxConfigBase::Get()->Read( wxT("/Misc/Misc/UseF10Key"), &USE_F10_KEY, true );
if ( USE_F10_KEY )
	gtk_settings_set_string_property(gtk_settings_get_default (), "gtk-menu-bar-accel", "F15", "foo");
#endif

Layout->m_notebook->LaunchFromMenu->LoadMenu(); 		// The menubar was loaded above from XRC.  This adds user-defined Launch commands to it

GetActivePane()->GetTreeCtrl()->SetFocus();						// If we don't do this, we start with NO focused window, which means some shortcuts fail

configure->ApplyTooltipPreferences();									// This doesn't work if it's done earlier!?
#if defined __WXGTK__
DnDStdCursor = wxCursor( BITMAPSDIR+wxT("DnDStdCursor.png") );	// Load the 2 cursors for DnD
DnDSelectedCursor = wxCursor( BITMAPSDIR+wxT("DnDSelectedCursor.png") );
#else
DnDStdCursor = wxCursor( wxCURSOR_NO_ENTRY );
DnDSelectedCursor = wxCursor( wxCURSOR_BULLSEYE);
#endif
m_dragMode = FALSE;															// Not DnDing
m_DragMutex.Lock();																// Start with it locked, so that OnBeginDrag can unlock it without asserting

#if defined __WXX11__ && wxVERSION_NUMBER < 2900
Connect( wxEVT_IDLE, (wxObjectEventFunction)&MyFrame::OnIdle );
#endif
}

#ifdef __WXX11__
void MyFrame::OnIdle( wxIdleEvent& WXUNUSED(event) )	// One-off to make the toolbar textctrl show the initial selection properly. Otherwise we only see the last 2 letters
{
if ( ! SelectionAvailable || GetActivePane() == NULL ) return;
wxString sel = GetActivePane()->GetPath(); GetActivePane()->DisplaySelectionInTBTextCtrl( sel );	// Show the selection again now the textctrl is big enough to cope!
Disconnect( wxID_ANY );
}
#endif

MyFrame::~MyFrame()
{
m_DragMutex.Unlock();

delete Layout;
delete Help;
delete AccelList;

configure->SaveConfiguration();
delete configure;
#if defined(__WXGTK20__)
PopEventHandler(true);															// In gtk2 we pushed a new event handler. Pop it to avoid a memory leak
#endif
}

void MyFrame::CreateMenuBar( wxMenuBar* MenuBar )
{
AccelList->FillMenuBar( MenuBar );
}

void MyFrame::RecreateMenuBar()
{
wxMenuBar* MenuBar = GetMenuBar();
AccelList->FillMenuBar( MenuBar, true );
}

void MyFrame::CreateAcceleratorTable()
{
int AccelEntries[] = {	SHCUT_LAUNCH_TERMINAL, SHCUT_TOOL_LOCATE, SHCUT_TOOL_FIND, SHCUT_TOOL_GREP, SHCUT_COMMANDLINE, SHCUT_TERMINAL_EMULATOR,
									SHCUT_ARCHIVE_EXTRACT, SHCUT_ARCHIVE_CREATE, SHCUT_ARCHIVE_COMPRESS, SHCUT_F1, SHCUT_CONFIG_SHORTCUTS, SHCUT_TOOLS_REPEAT
								};
const size_t shortcutNo = sizeof(AccelEntries)/sizeof(int);
AccelList->CreateAcceleratorTable( this, AccelEntries,  shortcutNo );
}

void MyFrame::ReconfigureShortcuts(wxNotifyEvent& WXUNUSED(event))
{
#if wxVERSION_NUMBER < 2600
	wxGetApp().RestartApplic();				// Since wxWidgets < 2.6.0 can't relabel menus properly, we have to restart 4Pane to make the changes happen
#endif

RecreateMenuBar();																						// First delete & recreate the menus
				// It would be nice to be able to empty the old accelerator table before replacing it, but trying caused segfaulting!
CreateAcceleratorTable();																				// Make a new one, with up-to-date data
															// Now recreate each pane's table
int count = Layout->m_notebook->GetPageCount(); if ( !count ) return; 		// Count the tabs
for ( int tab=0; tab < count; ++tab )																// For each tab
	{	MyTab* page = (MyTab*)Layout->m_notebook->GetPage( tab );			// Get ptr to it
		page->m_splitterLeftTop->m_left->RecreateAcceleratorTable();			// Tell each pane to recreate its accel table
		page->m_splitterLeftTop->m_right->RecreateAcceleratorTable();	
		page->m_splitterRightBottom->m_left->RecreateAcceleratorTable();
		page->m_splitterRightBottom->m_right->RecreateAcceleratorTable();
	}
}

void MyFrame::LoadToolbarButtons()	// Just to encapsulate loading tools onto the toolbar
{
wxBitmap toolBarBitmaps[7];

toolBarBitmaps[0] = wxBitmap(cut);
toolBarBitmaps[1] = wxBitmap(copy);
toolBarBitmaps[2] = wxBitmap(paste);
toolBarBitmaps[3] = wxBitmap(undo);
toolBarBitmaps[4] = wxBitmap(redo);
toolBarBitmaps[5] = wxBitmap(newtab);
toolBarBitmaps[6] = wxBitmap(deltab);

int count=0, separator=0;																				// Use these to count the no of buttons etc loaded before the ones with sidebars
		
																// Load the side-navigation buttons from XRC
largesidebar1 = (wxBitmapButton*)wxXmlResource::Get()->LoadObject(toolbar, wxT("IDM_TOOLBAR_LargeDropdown1"), wxT("wxBitmapButton"));
largesidebar2 = (wxBitmapButton*)wxXmlResource::Get()->LoadObject(toolbar, wxT("IDM_TOOLBAR_LargeDropdown2"), wxT("wxBitmapButton"));

toolbar->AddTool(SHCUT_CUT, _("Cut"), toolBarBitmaps[0], _("Click to Cut Selection") ); ++count;
toolbar->AddTool(SHCUT_COPY, _("Copy"), toolBarBitmaps[1], _("Click to Copy Selection") ); ++count;
toolbar->AddTool(SHCUT_PASTE, _("Paste"), toolBarBitmaps[2], _("Click to Paste Selection") ); ++count;
toolbar->AddSeparator(); ++separator;

buttonsbeforesidebar = count; separatorsbeforesidebar = separator;					// Store the counts, as here come the ones with sidebars
toolbar->AddTool(SHCUT_UNDO, _("UnDo"), toolBarBitmaps[3], _("Undo an action") );
toolbar->AddControl( largesidebar1 );
toolbar->AddTool(SHCUT_REDO, _("ReDo"), toolBarBitmaps[4], _("Redo a previously-Undone action") );
toolbar->AddControl( largesidebar2 );

toolbar->AddSeparator();
toolbar->AddTool(SHCUT_NEWTAB, _("New Tab"), toolBarBitmaps[5], _("Create a new Tab") );
toolbar->AddTool(SHCUT_DELTAB, _("Delete Tab"), toolBarBitmaps[6], _("Deletes the currently-selected Tab") );
		
toolbar->Realize();
}

void MyFrame::AddControls()				// Just to encapsulate adding textctrl & an editor button or 2 to the "toolbar", now actually a panel
{
wxBoxSizer* panelettesizer = new wxBoxSizer( wxHORIZONTAL );					// First set up a sizer for panelette
panelette->SetSizer( panelettesizer );

sizerControls = new wxBoxSizer( wxHORIZONTAL );										// This sizer will actually hold the device buttons
wxBoxSizer* sizerSpacer = new wxBoxSizer( wxHORIZONTAL );						// This one is just for a spacer, to prevent the buttons clustering at the extreme right
sizerSpacer->Add(5, 25, 0, wxRIGHT, 20); 														// Here's the spacer
EditorsSizer = new wxBoxSizer( wxHORIZONTAL );											// Put Editors etc here, so we can delete/reload them separately
FixedDevicesSizer = new wxBoxSizer( wxHORIZONTAL );								// Put the cdroms etc here, ditto

panelettesizer->Add( sizerControls, 1, wxEXPAND );
panelettesizer->Add( sizerSpacer, 0, wxEXPAND );

											// The first control to grab is (probably) the textctrl that holds the filepath of the selection
sizerControls->Add(20, 25);
m_tbText = new TextCtrlBase( panelette, ID_FILEPATH, wxT(""), false,  wxDefaultPosition, wxDefaultSize, wxTE_PROCESS_ENTER, wxDefaultValidator,  wxT("ID_FILEPATH") );
sizerControls->Add( m_tbText, 3,  wxALIGN_CENTER_VERTICAL );
if ( ! USE_DEFAULT_TERMINALEM_FONT )  m_tbText->SetFont( TERMINAL_FONT );	// It makes sense to use the same font as the terminal emulator
if ( m_tbText && ! SHOW_TB_TEXTCTRL )	m_tbText->Hide();						// I find it difficult to imagine anyone NOT wanting this, so I'll just hide it rather than doing anything cleverer
sizerControls->Add(20, 25);																				// Put a cosmetic spacer between this & the first editor

											// Now load the editor button(s)
DisplayEditorsInToolbar();

sizerControls->Add( EditorsSizer, 0, wxEXPAND );												// This is where the editor etc buttons will go
sizerControls->Add(10, 25, 0, wxEXPAND);														// Put a cosmetic spacer between the last editor and the drive buttons
sizerControls->Add( FixedDevicesSizer, 0, wxEXPAND );									// This is where the drive buttons will go
}

void MyFrame::DisplayEditorsInToolbar()		// These are stored in a separate sizer within the panelette sizer, for ease of recreation
{
						// We need to start by deleting any existing buttons, in case we're here from config & we're about to recreate
wxWindowList& panelchildren = panelette->GetChildren();
wxWindowListNode* node = panelchildren.GetLast();								// Start at the top, since we're deleting
while (node)																								// Iterate through the list of the sizer's children
	{	wxWindow* win = (wxWindow*)node->GetData();
	
		#if wxVERSION_NUMBER > 2503
		if ( EditorsSizer->GetItem(win) != NULL ) win->Destroy();					// If this window is found in the fixed-devices sizer, destroy it. If not, let it live
		#else
		wxList& children = EditorsSizer->GetChildren();									// If < 2.5.4 (when GetItem() first appeared), we need to provide a cut-down version of it
		wxWindowListNode* childnode = (wxWindowListNode*)children.GetFirst();
		while (childnode)
			{	wxSizerItem* item = (wxSizerItem*)childnode->GetData();
				if (item->GetWindow() == win)
					{ win->Destroy(); break; }

				childnode = childnode->GetNext();
			}
		#endif
		node = node->GetPrevious();
	}

size_t NoOfEditors = (size_t)wxConfigBase::Get()->Read(wxT("/Editors/NoOfEditors"), 0l);
for ( size_t n=0; n < NoOfEditors; ++n )
	{	EditorBitmapButton* button = new EditorBitmapButton( panelette, (size_t)n );
		if ( button->Ignore ) button->Destroy();												// If we don't want to know about this button, kill it
		 else EditorsSizer->Add( button, 0, wxEXPAND | wxRIGHT, 5 );		// Insert it into sizer
	}
}

void MyFrame::DoMiscUI(wxUpdateUIEvent& event)	//	UI for odds & ends
{
wxWindow* win = wxWindow::FindFocus(); if ( win == NULL ) return;
if ( win->GetName() == wxString(wxT("TerminalEm")) || win->GetName() == wxString(wxT("ID_FILEPATH")) )	// If the focus window is the TerminalEm etc
	 { event.Enable( event.GetId() == SHCUT_PASTE ); return; }	// All the other items are irrelevant here, and there's no easy way to know if there *is* anything relevant in the unix clipboard

MyGenericDirCtrl* active = GetActivePane(); bool IsArchive = (active!=NULL && active->arcman!=NULL && active->arcman->IsArchive());
switch( event.GetId() )								// Otherwise it's (presumably) a pane
	{	case SHCUT_PASTE:	{	bool clip = (MyGenericDirCtrl::filecount > 0); event.Enable( clip ); return; }												
		case SHCUT_RENAMETAB:
		case SHCUT_DELTAB:{	bool tab = (Layout->m_notebook->GetPageCount()); event.Enable( tab ); return; }
		case SHCUT_UNDO:	{ bool any = UnRedoManager::UndoAvailable(); event.Enable(any); largesidebar1->Enable(any); return; } // Enable Undo button & sidebar
		case SHCUT_REDO:	{ bool any = UnRedoManager::RedoAvailable(); event.Enable(any); largesidebar2->Enable(any); return; } // Enable Redo button & sidebar
		case SHCUT_NEW:	event.Enable( ! IsArchive ); return;
		case SHCUT_HARDLINK:
		case SHCUT_SOFTLINK:	event.Enable( ! IsArchive && (MyGenericDirCtrl::filecount > 0) ); return;	// We can't make new files or links within an archive
	}
}

void MyFrame::DoQueryEmptyUI(wxUpdateUIEvent& event)	//  This is for the items which are UIed depending on whether there's a selection
{
bool empty=false, DisableInArchive;
if ( ! SelectionAvailable ) empty = true;														// If there isn't a tab (or we're pretending so while we add/delete) 
  else 																											// There IS a valid tab.  Find the active pane, then ? anything selected
	{	wxWindow* win = wxWindow::FindFocus(); if ( win == NULL ) return;
		if ( win->GetName() == wxString(wxT("TerminalEm")) || win->GetName() == wxString(wxT("ID_FILEPATH")) )	// If the focus window is the TerminalEm etc
				empty = ( ((wxTextCtrl*)win)->GetStringSelection().IsEmpty() || event.GetId() > SHCUT_DELETE );

		 else
			{	MyGenericDirCtrl* active = GetActivePane();								// Otherwise it's (presumably) a pane
				DisableInArchive = (event.GetId()== SHCUT_PROPERTIES 	// We don't want to be able to ask about Properties inside an archive
															&& active!=NULL && active->arcman!=NULL && active->arcman->IsArchive());
				if ( active!=NULL && active->fileview == ISLEFT )	 empty = active->GetPath().IsEmpty();		// If a dirview this is straightforward
				 else 
					if ( active!=NULL && active->GetPath().IsEmpty() || active->GetPath() == active->startdir   // If fileview, path prob won't be empty even if no sel, so check if path==root
																											|| active->NoVisibleItems == true )	// or if the user has filtered out all the pane's content!
						empty = true;
			}
	}
event.Enable( ! (empty || DisableInArchive) );
}

void MyFrame::DoTabTemplateUI(wxUpdateUIEvent& event)	//  Grey when there are no Templates to load or delete
{
GetMenuBar()->Enable( event.GetId(), Layout->m_notebook->NoOfExistingTemplates ); 	// Do the Delete item, which is standard
	// Normally I'd have done the next item, LoadTemplate, in the same way, using EVT_UPDATE_UI_RANGE.
	// However, perhaps because it is a menu rather than an item, this won't work (the macro doesn't forward ANY events!?).  So piggyback on Delete
GetMenuBar()->Enable( LoadTabTemplateMenu, Layout->m_notebook->NoOfExistingTemplates );
}

void MyFrame::DoMiscTabUI(wxUpdateUIEvent& event)			//	UI for the tab-head items
{
if ( event.GetId()==SHCUT_ALWAYS_SHOW_TAB )  	
	GetMenuBar()->Check( SHCUT_ALWAYS_SHOW_TAB, Layout->m_notebook->AlwaysShowTabs );
 else
	GetMenuBar()->Check( SHCUT_SAME_TEXTSIZE_TAB, Layout->m_notebook->EqualSizedTabs );
}

void MyFrame::DoSplitPaneUI(wxUpdateUIEvent& event)
{
bool grey;
if ( GetActiveTab()==NULL ) grey = true;															// Disable menu item if no tab!
 else																												//  or if the orientation matches the menu-item command
 	{	if ( event.GetId()==SHCUT_SPLITPANE_VERTICAL )  	grey = (GetActiveTab()->splitstatus==vertical);
			else if ( event.GetId()==SHCUT_SPLITPANE_HORIZONTAL )  grey = (GetActiveTab()->splitstatus==horizontal);
			else if ( event.GetId()==SHCUT_SPLITPANE_UNSPLIT )	grey = (GetActiveTab()->splitstatus==unsplit);
	}

if ( event.GetId() >= SHCUT_SPLITPANE_VERTICAL && event.GetId() <= SHCUT_SPLITPANE_UNSPLIT )
			event.Enable( ! grey );
 else 	GetMenuBar()->Enable( event.GetId(), ! grey );
}

void MyFrame::DoColViewUI(wxUpdateUIEvent& event)
{
bool hidden;
FileGenericDirCtrl* filectrl;																				// We need a ptr to the filectrl of the active pane

MyGenericDirCtrl* view = GetActivePane(); if (view==NULL) return;
if ( view->fileview == ISLEFT )																		// If the left-panelet has focus, get its partner
	filectrl = (FileGenericDirCtrl*)view->partner;
 else filectrl = (FileGenericDirCtrl*)view;															// Otherwise it's the correct panelet already
 
hidden = filectrl->GetHeaderWindow()->IsHidden( event.GetId() - SHCUT_SHOW_COL_EXT + 1 );	 // Is this col showing or hidden?

event.Check( ! hidden );
}

void MyFrame::OnHelpF1(wxKeyEvent& WXUNUSED(event))
{
if ( HELPDIR.IsEmpty() ) return;																	// In case we couldn't find our helpfiles
switch( HelpContext )
	{	case HCconfigUDC:				Help->DisplaySection( wxT("ConfigureUserDefTools.htm") ); return;
		case HCconfigShortcuts:		Help->DisplaySection( wxT("ConfiguringShortcuts.htm") ); return;
		case HCconfigMisc:				Help->DisplaySection( wxT("ConfiguringMisc.htm") ); return;
		case HCconfigMiscExport:		Help->DisplaySection( wxT("Export.htm") ); return;
		case HCconfigMiscStatbar:	Help->DisplaySection( wxT("Statusbar.htm") ); return;
		case HCmultiplerename:			Help->DisplaySection( wxT("MultipleRenDup.htm") ); return;
		case HCconfigDevices:			Help->DisplaySection( wxT("ConfiguringDevices.htm") ); return;
		case HCconfigDisplay:			Help->DisplaySection( wxT("ConfiguringDisplay.htm") ); return;
		case HCconfigNetworks:		Help->DisplaySection( wxT("ConfiguringNetworks.htm") ); return;
		case HCconfigterminal:			Help->DisplaySection( wxT("ConfiguringTerminals.htm") ); return;
		case HCconfigMount:				Help->DisplaySection( wxT("Mount.htm") ); return;
		case HCarchive:						Help->DisplaySection( wxT("Archive.htm") ); return;
		case HCterminalEm:				Help->DisplaySection( wxT("TerminalEm.htm") ); return;
		case HCsearch:						Help->DisplaySection( wxT("Tools.htm") ); return;
		case HCproperties:					Help->DisplaySection( wxT("Properties.htm") ); return;
		case HCdnd:							Help->DisplaySection( wxT("DnD.htm") ); return;
		case HCbookmarks:				Help->DisplaySection( wxT("Bookmarks.htm") ); return;
		case HCfilter:							Help->DisplaySection( wxT("Filter.htm") ); return;
		case HCopenwith:					Help->DisplaySection( wxT("OpenWith.htm") ); return;
		case HCunknown:		
		 default:									Help->DisplaySection( wxT("Chapt.htm") ); return;
	}
}

void MyFrame::OnHelpContents(wxCommandEvent& WXUNUSED(event))
{
if ( HELPDIR.IsEmpty() ) return;																	// In case we couldn't find our helpfiles
Help->DisplaySection( wxT("Chapt.htm") );
}

void MyFrame::OnHelpFAQs(wxCommandEvent& WXUNUSED(event))
{
if ( HELPDIR.IsEmpty() ) return;																	// In case we couldn't find our helpfiles
Help->DisplaySection( wxT("FAQ.htm") );
}

void MyFrame::OnHelpAbout(wxCommandEvent& WXUNUSED(event))
{
if ( HELPDIR.IsEmpty() ) return;
HtmlDialog( _("About"), HELPDIR + wxT("/About.htm"), this, wxSize(300, 300) );
}

void MyFrame::OnSaveTabs(wxCommandEvent& WXUNUSED(event))
{
Layout->m_notebook->saveonexit = GetMenuBar()->IsChecked( SHCUT_SAVETABS_ONEXIT );	// Update the saveonexit bool, as this gets saved too
Layout->m_notebook->SaveDefaults(); 
}

void MyFrame::OnQuit(wxCommandEvent& WXUNUSED(event) )
{
Close(true);
}


void MyFrame::OnUpdateTrees(wxString path, wxArrayInt& PanesToUpdate, wxString newstartdir /*= NULL*/ )	// Tells all treectrls that branch 'path' needs updating
{																																															//  PanesToUpdate hold which panes MUST be done
static wxArrayInt IDs;
static wxArrayString PathsToUpdate;

int count = Layout->m_notebook->GetPageCount(); 					// How many tabs are there?
if ( !count ) return;

							// This method comes in 2 parts.  First we store the data, so that a cluster of actions can be updated all at once, not piecemeal
WX_APPEND_ARRAY( IDs, PanesToUpdate );										// Store the path and IDs in the appropriate static arrays
if ( path != wxT("**FLAG**") )																	// If this isn't a phantom path passed by twin OnUpdateTrees()
	{	if ( newstartdir.IsEmpty() )																// Check that we're not changing startdir 
			{	if (PathsToUpdate.Index( path ) == wxNOT_FOUND )				// Providing we don't have this path already
						PathsToUpdate.Add( path );											//		store it
			}
		 else
			{	PathsToUpdate.Add( wxT("**RECREATE**") );						// If we ARE changing startdir, flag this
				PathsToUpdate.Add( path );													//   & then store consecutively both path & newstartdir
				PathsToUpdate.Add( newstartdir );
			}
	}		 
	
if ( MyGenericDirCtrl::Clustering ) return;												// If the flag signals that we're in mid-cluster, do nothing else yet
	
							// Now for part 2.  If we're here, we've finished storing a cluster, so actually do the Updating
for ( size_t n=0; n < PathsToUpdate.Count(); ++n )									// For every stored item
	{	wxString path, newstartdir, parent;
		bool changeroot = false;
		
		path = PathsToUpdate[n];
		if ( path == wxT("**RECREATE**") )												// See if we've flagged to uproot a tree because of a change to 'root' item
			{	path = PathsToUpdate[++n]; newstartdir = PathsToUpdate[++n];	// If so, set it up
				if ( path != wxFILE_SEP_PATH ) parent = path.BeforeLast(wxFILE_SEP_PATH);	// Check path isn't root dir, as this would cause an error below
  					else parent = path;
				changeroot = true;
			}
		
		if ( !path.IsEmpty() )																		// Prune path of any terminal /'s to aid comparisons
			while ( path.Right(1) == wxFILE_SEP_PATH && path.Len() > 1 )  path.RemoveLast();
			
		for ( int tab=0; tab < count; ++tab )													// For each tab
			{	MyTab* page = (MyTab*)Layout->m_notebook->GetPage( tab );										// Get ptr to it
				DirGenericDirCtrl *dirL = page->m_splitterLeftTop->m_left;												// For sanity's sake, get ptrs to GenericDirCtrls
				FileGenericDirCtrl *dirLF = page->m_splitterLeftTop->m_right;	
				DirGenericDirCtrl *dirR = page->m_splitterRightBottom->m_left;
				FileGenericDirCtrl *dirRF = page->m_splitterRightBottom->m_right;
				if ( path==wxEmptyString )														// No path, because of error or need multiple alterations so easier just to RELOAD treectrls
					{	dirL->ReloadTree(); dirR->ReloadTree(); continue; }
				
				if ( changeroot )																		// We've renamed/destroyed startdir, so need to regrow at least one tree on different rootstock
					{	if ( dirL->startdir==path || dirL->startdir.BeforeLast(wxFILE_SEP_PATH)==path )		// If L dir-pane has path as its 'root'
							{	dirL->startdir = newstartdir; dirL->ReCreateTreeFromSelection(); dirLF->ReCreateTreeFromSelection(); }	//   redo it from new startdir
						 else																				// Otherwise, refresh but with parent as path. This catches panes that contain path but not as 'root'
							{	dirL->ReloadTree( parent, IDs ); dirLF->ReloadTree( parent, IDs ); }
							
						if ( dirR->startdir==path || dirR->startdir.BeforeLast(wxFILE_SEP_PATH)==path )	// Ditto for R pane
							{	dirR->startdir = newstartdir; dirR->ReCreateTreeFromSelection();  dirRF->ReCreateTreeFromSelection();}
						 else
							{	dirR->ReloadTree( parent, IDs ); dirRF->ReloadTree( parent, IDs ); }
					}
				 else
					{	if ( dirL->startdir==path || dirL->startdir.BeforeLast(wxFILE_SEP_PATH)==path )	// Again check if we're refreshing startdir, even if it hasn't changed
							{ dirL->ReCreateTreeFromSelection(); dirLF->ReCreateTreeFromSelection(); }
						 else { dirL->ReloadTree( path, IDs ); dirLF->ReloadTree( path, IDs ); }					// Otherwise, do standard Refreshes
						if ( dirR->startdir==path || dirR->startdir.BeforeLast(wxFILE_SEP_PATH)==path )
							{ dirR->ReCreateTreeFromSelection(); dirRF->ReCreateTreeFromSelection(); }
						 else { dirR->ReloadTree( path, IDs ); dirRF->ReloadTree( path, IDs ); }
					}
			}
	}
	
PathsToUpdate.Clear(); IDs.Clear();														// Clear the arrays, ready for next time
}

void MyFrame::UpdateTrees()	// At the end of a cluster of actions, tells all treectrls actually to do the updating they've stored
{

wxString path(wxT("**FLAG**")); wxArrayInt IDs;									// Create 'empty' parameters to pass (wxEmptyString is already in use)

MyGenericDirCtrl::Clustering = false;													// Reset the flag that was preventing the updating

OnUpdateTrees( path, IDs );																	// Now call its twin, using blank parameters.  The reset flag means things actually happen
}

void MyFrame::OnUnmountDevice( wxString& path )	// When we've successfully umounted eg a floppy, reset treectrl(s) displaying it to HOME
{
if ( path==wxEmptyString )	 return;

int count = Layout->m_notebook->GetPageCount(); 							// How many tabs are there?
if ( !count ) return;

wxString newdir, chosen;
newdir = HOME;																						// Use this as default

wxConfigBase *config = wxConfigBase::Get();										// Find the config data, or load it if not already done
if (config==NULL)  { wxLogError(_("Couldn't load configuration!?")); return; }

long choice; config->Read( wxT("/DefaultPathOnUmount/Type"), &choice );	// Find out which case to use below

switch( choice )
	{	case current:		newdir = wxGetCwd(); break;								// One alternative is to use the current working dir
		case selected:	config->Read( wxT("/DefaultPathOnUmount/Chosen"), &chosen ); 		// Another is to use choice stored in config. 
									wxFileName fn( chosen );																// If so, first check it's a valid dir!
									if ( wxDirExists( fn.GetFullPath()) )  newdir = chosen;						// If it is, use it.  If not, stick to default
	}
	
for ( int tab=0; tab < count; tab++ )														// For each tab
	{	MyTab* page = (MyTab*)Layout->m_notebook->GetPage( tab );									// For sanity's sake, get ptrs to GenericDirCtrls
		DirGenericDirCtrl *dirL = page->m_splitterLeftTop->m_left;			
		DirGenericDirCtrl *dirR = page->m_splitterRightBottom->m_left;

		if ( dirL->startdir == path )		dirL->OnOutsideSelect( newdir );
		if ( dirR->startdir == path )	dirR->OnOutsideSelect( newdir );
	}
}

void MyFrame::OnUndo(wxCommandEvent& WXUNUSED(event))
{
UnRedoManager::UndoEntry();								// Do the Undoable
}

void MyFrame::OnUndoSidebar(wxCommandEvent& WXUNUSED(event))
{
UnRedoManager::OnUndoSidebar();						// Do the multiply-Undoable	
}

void MyFrame::OnRedo(wxCommandEvent& WXUNUSED(event))
{
UnRedoManager::RedoEntry();								// ReDo the Redoable
}

void MyFrame::OnRedoSidebar(wxCommandEvent& WXUNUSED(event))
{
UnRedoManager::OnRedoSidebar();
}

void MyFrame::OnCut(wxCommandEvent& event)
{
GetActivePane()->OnShortcutCut( event );											// Find the active MyGenericDirCtrl & pass request to it
}

void MyFrame::OnCopy(wxCommandEvent& event)
{
GetActivePane()->OnShortcutCopy( event );
}

void MyFrame::OnPaste(wxCommandEvent& event)
{
GetActivePane()->OnShortcutPaste( event );
}

void MyFrame::OnNew(wxCommandEvent& event)
{
GetActivePane()->NewFile( event );
}

void MyFrame::OnHardLink(wxCommandEvent& event)
{
GetActivePane()->OnShortcutSoftLink(event);
}

void MyFrame::OnSoftLink(wxCommandEvent& event)
{
GetActivePane()->OnShortcutSoftLink(event);
}

void MyFrame::OnTrash(wxCommandEvent& event)
{
wxWindow* win = wxWindow::FindFocus();												// This should return the TreeCtrl
if ( win == NULL ) { wxLogDebug( wxT("FindFocus gave a null window") ); return; }
if ( win->GetName() == wxString(wxT("MyTreeCtrl")) )								// If it's a treectrl
	((MyGenericDirCtrl*)(win->GetParent()))->OnShortcutTrash( event);	// The parent will be the relevant MyGenericDirCtrl
}

void MyFrame::OnDelete(wxCommandEvent& event)
{
wxWindow* win = wxWindow::FindFocus();											// This should return the TreeCtrl
if ( win == NULL ) { wxLogTrace( wxT("%s"),wxT(" FindFocus gave a null window") ); return; }
if ( win->GetName() == wxString(wxT("MyTreeCtrl")) )							// If it's a treectrl
	((MyGenericDirCtrl*)(win->GetParent()))->OnShortcutDel( event);	// The parent will be the relevant MyGenericDirCtrl
}

void MyFrame::OnRename(wxCommandEvent& event)
{
GetActivePane()->OnRename(event);													// Find the active MyGenericDirCtrl & pass request to it
}

void MyFrame::OnDup(wxCommandEvent& event)
{
GetActivePane()->OnDup(event);
}

void MyFrame::OnProperties(wxCommandEvent& event)
{
GetActivePane()->OnProperties(event);
}

void MyFrame::OnFilter(wxCommandEvent& event)
{
GetActivePane()->OnFilter(event);
}

void MyFrame::OnToggleHidden(wxCommandEvent& event)
{
MyGenericDirCtrl* active = GetActivePane();
if ( active==NULL )		return;
if ( active->fileview == ISRIGHT ) ((FileGenericDirCtrl*)active)->OnToggleHidden(event);
	else ((DirGenericDirCtrl*)active)->OnToggleHidden(event);
}

void MyFrame::OnArchiveExtract(wxCommandEvent& WXUNUSED(event))
{
MyGenericDirCtrl* active = GetActivePane();
#if wxVERSION_NUMBER >= 2504
if ( active->arcman->IsArchive() )				// If we're peering into an archive stream, tell this what to do
	{ active->arcman->GetArc()->OnExtract(); return; }
#endif
Archive* archive = new Archive( active );	// Otherwise do the standard-type extract
HelpContext = HCarchive;
archive->ExtractOrVerify();
HelpContext = HCunknown;
}

void MyFrame::OnArchiveCreate(wxCommandEvent& WXUNUSED(event))
{
Archive* archive = new Archive( GetActivePane() );
HelpContext = HCarchive;
archive->CreateOrAppend();																// This calls CreateOrAppend() with a default signifying create
HelpContext = HCunknown;
}

void MyFrame::OnArchiveAppend(wxCommandEvent& WXUNUSED(event))
{
Archive* archive = new Archive( GetActivePane() );
HelpContext = HCarchive;
archive->Append();																			// This calls CreateOrAppend() with a bool signifying append
HelpContext = HCunknown;
}

void MyFrame::OnArchiveTest(wxCommandEvent& WXUNUSED(event))
{
Archive* archive = new Archive( GetActivePane() );
HelpContext = HCarchive;
archive->ExtractOrVerify( false );														// False signifies Verify
HelpContext = HCunknown;
}

void MyFrame::OnArchiveCompress(wxCommandEvent& WXUNUSED(event))
{
Archive* archive = new Archive( GetActivePane() );
HelpContext = HCarchive;
archive->Compress();
HelpContext = HCunknown;
}

void MyFrame::OnLaunchTerminal(wxCommandEvent& WXUNUSED(event))		// Ctrl-T by default.  Launch a stand-alone terminal session, using the selected dir 
{
wxString path;
wxString oldwd = wxGetCwd();															// Save the Current working directory, to restore later.  ?necessary, but it sounds like good practice

path = GetActivePane()->GetPath();													//  Get the dir currently highlit. Path, not filepath
if ( !path.IsEmpty() )
	{	FileData fd( path ); if ( ! fd.IsDir() ) path = fd.GetPath();			// because wxSetWorkingDirectory() borks if it's passed a FILEpath
		wxSetWorkingDirectory(path);													// If it's valid, make it the current working dir, so that the terminal will cd into this
	}
	
wxExecute( TERMINAL );																	// Launch the terminal session
wxSetWorkingDirectory(oldwd);															// Return to the original wd
}

void MyFrame::OnCommandLine(wxCommandEvent& WXUNUSED(event))
{
Layout->OnCommandLine();
}

void MyFrame::OnMountPartition(wxCommandEvent& WXUNUSED(event))
{
HelpContext = HCconfigMount;
Layout->m_notebook->DeviceMan->OnMountPartition();
HelpContext = HCunknown;
}

void MyFrame::OnUnMountPartition(wxCommandEvent& WXUNUSED(event))
{
HelpContext = HCconfigMount;
Layout->m_notebook->DeviceMan->OnUnMountPartition();
HelpContext = HCunknown;
}

void MyFrame::OnMountIso(wxCommandEvent& WXUNUSED(event))
{
HelpContext = HCconfigMount;
Layout->m_notebook->DeviceMan->OnMountIso();
HelpContext = HCunknown;
}

void MyFrame::OnMountNfs(wxCommandEvent& WXUNUSED(event))
{
HelpContext = HCconfigMount;
Layout->m_notebook->DeviceMan->OnMountNfs();
HelpContext = HCunknown;
}

void MyFrame::OnMountSamba(wxCommandEvent& WXUNUSED(event))
{
HelpContext = HCconfigMount;
Layout->m_notebook->DeviceMan->OnMountSamba();
HelpContext = HCunknown;
}

void MyFrame::OnUnMountNetwork(wxCommandEvent& WXUNUSED(event))
{
HelpContext = HCconfigMount;
Layout->m_notebook->DeviceMan->OnUnMountNetwork();
HelpContext = HCunknown;
}

void MyFrame::OnTermEm(wxCommandEvent& WXUNUSED(event))
{
Layout->DoTool( terminalem );
}

void MyFrame::OnLocate(wxCommandEvent& WXUNUSED(event))
{
Layout->DoTool( locate );
}

void MyFrame::OnFind(wxCommandEvent& WXUNUSED(event))
{
Layout->DoTool( find );
}

void MyFrame::OnGrep(wxCommandEvent& WXUNUSED(event))
{
Layout->DoTool( grep );
}

void MyFrame::OnToolsLaunch(wxCommandEvent& event)
{
int id = event.GetId();
if ( id < SHCUT_TOOLS_LAUNCH || id >= SHCUT_TOOLS_LAUNCH_LAST )
 		return;																						// Out of range so abort
	
Layout->m_notebook->LaunchFromMenu->Run( id );							// Pass to the relevant class
}

void MyFrame::OnToolsRepeat(wxCommandEvent& WXUNUSED(event))	// Repeat the last command
{
Layout->m_notebook->LaunchFromMenu->RepeatLastCommand();
}

void MyFrame::DoRepeatCommandUI(wxUpdateUIEvent& event)	//  Disable when there hasn't been a previous command to repeat
{
GetMenuBar()->Enable( event.GetId(), (int)Layout->m_notebook->LaunchFromMenu->GetLastCommand() != -1 ); 
}

void MyFrame::OnEmptyTrash(wxCommandEvent& event)
{
Layout->m_notebook->DeleteLocation->EmptyTrash( event.GetId() == SHCUT_EMPTYTRASH );	// Either empty the trash-can or the 'deleted'can
}

void MyFrame::ToggleFileviewSizetype(wxCommandEvent& WXUNUSED(event))
{
SHOW_RECURSIVE_FILEVIEW_SIZE = ! SHOW_RECURSIVE_FILEVIEW_SIZE;
}

void MyFrame::ToggleRetainRelSymlinkTarget(wxCommandEvent& WXUNUSED(event))
{
RETAIN_REL_TARGET = ! RETAIN_REL_TARGET;
}

void MyFrame::OnConfigure(wxCommandEvent& WXUNUSED(event))
{
configure->Configure4Pane();
}

void MyFrame::OnBeginDrag()
{
if ( m_DragMutex.TryLock() == wxMUTEX_NO_ERROR ) wxLogTrace( wxT("%s"),wxT("Entering OnBeginDrag:  Mutex was UNLOCKED!?"));	// Temp test
if (m_dragMode == true)      return;														// Can't imagine how this could happen, but . . .

wxPoint pt = ScreenToClient( wxGetMousePosition() );						// Where's the mouse?  Screen coords so -> client

														// The next bit is taken from wx/generic/dragimgg.h
#ifdef __WXMSW__
    ::UpdateWindow((HWND) GetHWND()); // We need to implement this in wxWindows
#endif

oldCursor = GetCursor();																	// Grab the current cursor
SetCursor( DnDStdCursor );																//  & alter it for the duration

m_dragImage = new MyDragImage();
if ( !m_dragImage ) { wxLogTrace( wxT("%s"),wxT("m_dragImage creation failed")); SetCursor( oldCursor ); return; }

m_dragImage->pt = pt;
m_dragImage->SetImageType( ParseMetakeys() );								// This sets the initial image type according to the metakey state - - - drag, softlink etc

m_dragImage->BeginDrag( wxPoint(0, 0), this, true );							// the 'true' means use fullscreen mode.  Needed, as otherwise the image overwrote the statusbar!?
m_dragImage->Move(pt);
m_dragImage->Show();

lastwin = wxFindWindowAtPointer( pt );												// Store which treectrl we're dealing with

m_dragMode = true;																			// Flags that we're in DnD mode
m_DrawImageisShown = true;															// & image isn't hidden
m_LeftUp = false;															
m_finished = false;
ESC_Pressed = false;
m_DragMutex.Unlock();																		// Finally, do the big unlock that makes everything work
}

void MyFrame::DoMutex(wxMouseEvent& event)
{
if ( m_DragMutex.TryLock() == wxMUTEX_NO_ERROR ) 						// If the mutex is free, lock it & pass the event on to the correct method
	{ if ( event.LeftUp() || m_LeftUp )		OnEndDrag(event);					//  ie end if LeftUp now or recently
			else
				{ if ( event.Dragging() )  OnMouseMove(event);					//   otherwise move
						else  { m_DragMutex.Unlock(); return; }						// If all else fails,  abort
				}
	}
 else
	{ if ( event.LeftUp() && ! m_finished )	m_LeftUp = true;					// If we just caught LeftUp, & it's the real one, flag it for next time
		return;																						// Either way, abort
	}
}

void MyFrame::OnESC_Pressed()	// If we're in DnD, this will cancel it.  If not, nothing
{
if ( m_dragMode == true )
	{ ESC_Pressed = true; m_LeftUp = true; }										// This means that the next time thru, DoMutex will call OnEndDrag, which will abort
}


void MyFrame::OnMouseMove(wxMouseEvent& event)
{
if (m_dragMode == false) { wxLogTrace( wxT("%s"),wxT("m_dragMode == FALSE")); return; }

// Prevent screen corruption by hiding the image		************ Not needed, and it's MUCH faster without
//if (m_dragImage)		m_dragImage->Hide();

#ifdef __WXMSW__
if (updateWindow)
		::UpdateWindow((HWND) GetHWND());
#endif

wxPoint pt;
wxWindow *win = wxFindWindowAtPointer( pt );								// Find window under mouse, & pointer screen coords

if ( win == NULL ) { m_DragMutex.Unlock(); return; }						// Otherwise it crashes, when outside this frame

if ( win != lastwin )																				// If we've left the previous window
	{	m_dragImage->SetCursor( dndUnselect );									// In case cursor is the Selected type, revert to standard
		if ( lastwin->GetName() == wxT("MyTreeCtrl") )							// If we just left a treectrl
				m_dragImage->OnLeavingPane( lastwin ); 						//  		tell MyDragImage, which may start the scroll-timer
		lastwin = win; 																				// Store current window	
	}

if ( win->GetName() == wxT("MyTreeCtrl") )										// If we're within a treectrl
	{	
		MyTreeCtrl* tree = (MyTreeCtrl*)win;											// For convenience
	if ( tree==NULL ) wxLogTrace( wxT("%s"),wxT("*************** tree was dead wood! *******************"));
		pt = win->ScreenToClient( pt );													// Translate coords
		
//		tree->OnDragOver( pt );					// Use the win to emphasise which treeitem (if any) is currently pointed at
// I couldn't get the above to work without segfaulting every so often.  The problem is that doing something to the tree window requires a 'refresh' before
// it displays.  This doesn't happen immediately, and occasionally by the time it DOES, the tree control has been destroyed, so making the emphasis happen to
// a null ptr.  Result, segfault.  There is also a problem with the emphasis refresh overwriting the dragimage (since we're not hiding/showing).  All in all, not worth it!

		int flags; wxTreeItemId item = tree->HitTest( pt, flags );				// Find out if we're near a tree item
		if ( item.IsOk() && !(flags & wxTREE_HITTEST_ONITEMRIGHT) )
				m_dragImage->SetCursor( dndSelectedCursor );
		  else  m_dragImage->SetCursor( dndStdCursor );			
    }
 else if ( win->GetName() == wxT("Editor") || win->GetName() == wxT("Device") || win->GetName() == wxT("TerminalEm") )
			m_dragImage->SetCursor( dndSelectedCursor );

m_dragImage->pt = ClientToScreen( event.GetPosition() );					// Store the mouse position, so that the animation timer can redraw the image correctly
m_CtrlPressed = event.m_controlDown;													// Store the pattern of metakeys, so we know if we're copying, moving etc 
m_ShPressed = event.ShiftDown();
m_AltPressed = event.AltDown();
m_dragImage->SetImageType( ParseMetakeys() );								// Change image type to fit the metakey state - - - drag, softlink etc

				// Move and show the image again		************ I killed the Show, not needed
if ( m_dragMode ) m_dragImage->Move(event.GetPosition()); 				// Now do the Image Move
	else  return;

		// We need to Yield() from time to time, otherwise other events accumulate which have to run their course before the Drop, resulting in a second or 2 of freeze
		// However Yielding every mouse-movement makes the Drag very jerky, so do it 1:10
static size_t counter = 0;
counter++; if ( ! (counter % 10) ) wxYieldIfNeeded();
		
if ( m_LeftUp && !m_finished ) { OnEndDrag(event); return; }				// If LeftUp happened while we were here, jump to EndDrag

m_DragMutex.Unlock();
}

void MyFrame::OnEndDrag(wxMouseEvent& event) 
{
if (m_dragMode == FALSE)																// Check if we're in DnD or not
	{ m_TreeDragMutex.Unlock(); return; }
m_finished = TRUE;																			// Set the flag that prevents reentry

m_dragMode = FALSE;																		// Unflag DnD

// End the drag and delete the drag image
if (m_dragImage)
	{	m_dragImage->Hide();																// Shouldn't be necessary, but it is.
		m_dragImage->EndDrag();
		delete m_dragImage;
		m_dragImage = NULL;
	}

SetCursor(oldCursor);

if ( ESC_Pressed ) { m_TreeDragMutex.Unlock(); return; }					// If we're here because of an ESC, abort

								// Finally, we need to hand over to the destination window to do the actual file-work 
wxPoint pt;
wxWindow *win = wxFindWindowAtPointer( pt );									// Find window under mouse, & pointer screen coords
if ( win != NULL ) 																				// (Otherwise it crashes when LButtonUp occurs outside this frame) 
	{	if ( win->GetName() == wxT("MyTreeCtrl") ) 								// Is it a tree . . . . ?						
			{	((MyTreeCtrl*)win)->OnEndDrag((wxTreeEvent&)event); return; }
		if ( win->GetName() == wxT("Editor") )										// Is it an editor . . . . ?
			{	((EditorBitmapButton*)win)->OnEndDrag(); return; }	
		if ( win->GetName() == wxT("Device") )										// Is it a device . . . . ?
					{	((DeviceBitmapButton*)win)->OnEndDrag(); return; }
		if ( win->GetName() == wxT("TerminalEm") )								// Are we dropping onto the Terminal Emulator or Command Line ?
					{	((TerminalEm*)win)->OnEndDrag(); return; }
	}

m_TreeDragMutex.Unlock();
}

void MyFrame::OnDnDRightclick(wxMouseEvent& event)
{
if (m_dragMode == FALSE) { event.Skip(); return; }								// We only want it to happen during D'n'D

m_dragImage->DoPageScroll( event );
}

void MyFrame::OnMouseWheel( wxMouseEvent& event )
{
if (m_dragMode == FALSE) { event.Skip(); return; }								// We only want it to happen during D'n'D

m_dragImage->DoLineScroll( event );
}

enum myDragResult MyFrame::ParseMetakeys()	// Use the metakey store to find if we're trying to copy, move, link . . .  
{
int metakeys = wxACCEL_NORMAL;
if ( wxGetKeyState(WXK_CONTROL) )  metakeys |= wxACCEL_CTRL;
if ( wxGetKeyState(WXK_SHIFT) )  metakeys |= wxACCEL_SHIFT;
if ( wxGetKeyState(WXK_ALT) )  metakeys |= wxACCEL_ALT;

if  ( metakeys == METAKEYS_COPY ) return myDragCopy;
if  ( metakeys == METAKEYS_HARDLINK ) return myDragHardLink;
if  ( metakeys == METAKEYS_SOFTLINK ) return myDragSoftLink;
return myDragMove;																				// Seems a sensible default
}


MyTab* MyFrame::GetActiveTab()	// Returns the currently-active tab, if there is one
{
if (Layout->m_notebook==NULL) 	return NULL;
if ( Layout->m_notebook->GetPageCount() == 0 ) 	return NULL;

int current = Layout->m_notebook->GetSelection(); 								// Get current page-no
if ( current == wxNOT_FOUND ) return NULL;										// Shouldn't happen, since we checked there are >0 pages

return (MyTab*)( Layout->m_notebook->GetPage( current ) );				// Return the page, or null if it doesn't exist
}

void MyFrame::RefreshAllTabs( wxFont* font, bool ReloadDirviewTBs/*=false*/  )	// Refreshes all panes on all tabs, optionally with a new font, or mini-toolbars
{
if (Layout->m_notebook==NULL) 	return;
size_t count = Layout->m_notebook->GetPageCount(); if ( ! count ) 	return;

for ( size_t n=0; n < count; ++n )
	{	MyTab* tab = (MyTab*)Layout->m_notebook->GetPage( n ); if ( tab == NULL || tab->LeftorTopDirCtrl == NULL )  continue;
		if ( ReloadDirviewTBs )  tab->LeftorTopDirCtrl->ReloadDirviewToolbars();
		 else tab->LeftorTopDirCtrl->RefreshVisiblePanes( font );
	}
}

void MyFrame::SetSensibleFocus()		// Ensures that keyboard focus is held by a valid window, usually the active pane
{
HelpContext = HCunknown;
MyGenericDirCtrl* active = GetActivePane();
if ( active ) { active->GetTreeCtrl()->SetFocus(); return; }				// The standard scenario	

if ( m_tbText && m_tbText->IsShown() )
	{ m_tbText->SetFocus(); return; }													// If all the tabs have been deleted, try for the toolbar textctrl

SetFocus();																							// Failing all else, use the frame itself
}

MyGenericDirCtrl* MyFrame::GetActivePane()		// Returns the most-recently-active pane in the visible tab
{
MyTab* tab = GetActiveTab();
if ( tab==NULL) 	return NULL;

return tab->LeftRight;																		// Return the answer, which points to one of the panes
}

void MyFrame::OnReplicate(wxCommandEvent& event){ GetActivePane()->OnReplicate(event); }			// Find the active MyGenericDirCtrl & pass request to it
void MyFrame::OnSwapPanes(wxCommandEvent& event){ GetActivePane()->OnSwapPanes(event); }	// Ditto


void MyFrame::OnAddToBookmarks(wxCommandEvent& WXUNUSED(event))
{
wxString path = GetActivePane()->GetPath();									// Find the current selection, which is presumably to be bookmarked
HelpContext = HCbookmarks;
Layout->m_notebook->BookmarkMan->AddBookmark(path);			//  & pass it to the Bookmark manager
HelpContext = HCunknown;
}

void MyFrame::OnManageBookmarks(wxCommandEvent& WXUNUSED(event))
{
HelpContext = HCbookmarks;
Layout->m_notebook->BookmarkMan->ManageBookmarks();
HelpContext = HCunknown;
}

void MyFrame::OnBookmark(wxCommandEvent& event)
{
wxString filepath, path = Layout->m_notebook->BookmarkMan->RetrieveBookmark( event.GetId() );	// Use Bookmark manager to retrieve the path of the requested bookmark
if ( path.IsEmpty() )  return;
filepath = path;																					// Store it, as it gets altered in OnOutsideSelect()

MyGenericDirCtrl* active = GetActivePane();
if ( active==NULL )		return;
if ( active->fileview == ISRIGHT )  active = active->partner;				// It's much easier if we use the dirview to set both panes

active->OnOutsideSelect( filepath );													// If everything's valid, select it into the currently-active dirview

if ( active->fulltree )																				//	If in fulltree mode, make fileview-pane match too.  If not, OnOutsideSelect() does it anyway
	{	active->partner->startdir = path;	
		active->partner->ReCreateTreeFromSelection();
	}

FileData fd( path );																				// Is it a dir or a file?  If it's a file, it's nice to set the filepath to it in the fileview
if ( fd.IsDir() ) active->SetPath( path );
 else if ( fd.IsValid() || active->arcman->IsArchive() ) active->partner->SetPath( path );

active->DisplaySelectionInTBTextCtrl( path );										//	Put it in the textctrl too
active->GetTreeCtrl()->SetFocus();													// Clicking the menubar took the focus from the pane.  This returns it
}

void MyFrame::TextFilepathEnter()		// The Filepath textctrl has the focus and Enter has been added, so GoTo its contents
{
if ( m_tbText == NULL ) return;															// Not very likely if we managed to get here!
wxString filepath, path = m_tbText->GetValue();								// Get the path requested
if ( path.IsEmpty() ) 	return;
filepath = path;																					// Store it, as it gets altered in OnOutsideSelect()

MyGenericDirCtrl* active = GetActivePane();
if ( active==NULL )		return;
if ( active->fileview == ISRIGHT )  active = active->partner;				// It's much easier if we use the dirview to set both panes

active->OnOutsideSelect( filepath );													// If everything's valid, select it into the currently-active dirview

FileData fd( path );																				// We're going to need this in a minute, so use it to get the path cf filepath
if ( active->fulltree )																				//	If in fulltree mode, make fileview-pane match too.  If not, OnOutsideSelect() does it anyway
	{	active->partner->startdir = fd.GetPath();	
		active->partner->ReCreateTreeFromSelection();
	}

if ( fd.IsDir() ) active->SetPath( path );												// Is it a dir or a file?  If it's a file, it's nice to set the filepath to it in the fileview
 else active->partner->SetPath( path );
}

void MyFrame::SplitHorizontal( wxCommandEvent& WXUNUSED(event))
{
MyTab* tab = GetActiveTab();															// Find the active tab
if (tab==NULL) 	return;

tab->PerformSplit( horizontal );															// Tell it how to split itself
}

void MyFrame::SplitVertical(wxCommandEvent& WXUNUSED(event))
{
MyTab* tab = GetActiveTab();
if (tab==NULL) 	return;

tab->PerformSplit( vertical );
}

void MyFrame::Unsplit(wxCommandEvent& WXUNUSED(event))
{
MyTab* tab = GetActiveTab();
if (tab==NULL) 	return;

tab->PerformSplit( unsplit );
}

void MyFrame::OnTabTemplateLoadMenu(wxCommandEvent& event)
{
Layout->m_notebook->LoadTemplate( event.GetId() - SHCUT_TABTEMPLATERANGE_START );
}

void MyFrame::OnConfigureShortcuts(wxCommandEvent& WXUNUSED(event))
{
AccelList->ConfigureShortcuts();
}

//--------------------------------------------------------------------------------------------------------------------------------------
LayoutWindows::LayoutWindows(MyFrame *parent)		:	m_frame( parent )
{
tool = NULL; CommandlineShowing = EmulatorShowing = false;
bottompanel = NULL;
}

LayoutWindows::~LayoutWindows()
{
wxConfigBase::Get()->Write(wxT("/Misc/PropertiesPage"), m_frame->PropertiesPage );	// Save the last-used page of the Properties notebook

m_notebook->saveonexit = m_frame->GetMenuBar()->IsChecked( SHCUT_SAVETABS_ONEXIT );
if ( m_notebook->saveonexit ) m_notebook->SaveDefaults();				// If we're supposed to save tabdata etc, do so

if ( tool != NULL ) delete tool; 
}

void LayoutWindows::Setup()
{
		// Start with the first division: a textctrl with a standard rest-of-screen wxSplitterWindow below, using a sizer not a splitter to position
commandlinepane = wxXmlResource::Get()->LoadPanel( m_frame, wxT("CommandPane") );	// Load the pane for displaying the command-line
m_frame->sizerMain->Add( commandlinepane, 0, wxEXPAND );
m_frame->sizerMain->Show(commandlinepane, false);
m_frame->sizerMain->Layout();

																			// Here comes the rest-of-screen splitter-window
m_splitterMain = new wxSplitterWindow(m_frame, -1, wxDefaultPosition, wxDefaultSize,  0);
m_frame->sizerMain->Add( m_splitterMain, 1, wxEXPAND, 0 );
m_frame->SetSizer( m_frame->sizerMain );

		// Right, we want something to put in the main splitterwindow: a Notebook to hold the secondary splitter, and something below to hold terminal sessions etc
m_notebook = new MyNotebook( m_splitterMain, NextID++ );
m_splitterMain->Initialize(m_notebook);											// Initialise the notebook part of m_splitterMain

		// Load the preferred series of tabs, or else a default one
m_notebook->LoadTabs( -1, ((MyApp&)wxGetApp()).startdir0, ((MyApp&)wxGetApp()).startdir1 );	// The 2nd/3rd parameters are in case the user typed 4Pane /path/to/foo, /path/to/bar

														// While we're here, load which page of the Properties notebook to start with
wxConfigBase::Get()->Read( wxT("/Misc/PropertiesPage"), &m_frame->PropertiesPage, 0 );

	// Now show the commandline (which calls the terminal emulator too) or the terminal emulator, if requested
if ( m_notebook->showcommandline )		OnCommandLine();
  else if ( m_notebook->showterminal )	DoTool( terminalem );
}

void LayoutWindows::OnChangeDir( wxString& newdir, bool ReusePrompt )	// Called when a terminalem does a cd, to tell commandline or vice versa
{																// Also called when a dirview changes selection, so that this can be reflected in any terminal/commandline prompt
if ( EmulatorShowing ) 
	{	tool->text->OnChangeDir( newdir, ReusePrompt );
		if ( CommandlineShowing ) commandline->WritePrompt();		// If the emulator's there, see if the commandline is too 
	}
}

void LayoutWindows::OnCommandLine()
{
if ( ! CommandlineShowing )															// If it's not showing, load & show the command-line
	{	DoTool( cmdline );
		ShowHideCommandLine();														// If it's showing, hide it; & vice versa
	}
 else		// NB tool->Close() will itself call ShowHideCommandLine()
 	tool->Close();																				// If it IS showing, shut it down, closing the bottom panel in the process
}

void LayoutWindows::ShowHideCommandLine()			// If it's showing, hide it; & vice versa
{
CommandlineShowing = ! CommandlineShowing;							// Negate the bool ready for next time
m_frame->sizerMain->Show(commandlinepane, CommandlineShowing);	// Show/Hide appropriately
m_frame->sizerMain->Layout();
}

bool LayoutWindows::ShowBottom()		// If bottompanel isn't visible, show it
{
if ( m_splitterMain->IsSplit() ) return false;									// Can't split it into four

#if wxVERSION_NUMBER >=2500
if ( bottompanel==NULL )					// >2.4.2 we Detach() from the sizer each time it's closed. Beforehand it was Removed, so we always need to recreate it
#endif
	bottompanel = wxXmlResource::Get()->LoadPanel( m_splitterMain, wxT("TerminalPane") );

int w, h; m_notebook->GetClientSize( &w, &h );
m_splitterMain->SplitHorizontally(m_notebook, bottompanel, h *2/3 );
return true;
}

void LayoutWindows::UnShowBottom()		// If bottompanel is visible, unsplit it 
{
if ( ! m_splitterMain->IsSplit() ) return;											// Can't unsplit it if it's not split

m_splitterMain->Unsplit(bottompanel);
}

void LayoutWindows::DoTool( enum toolchoice whichtool )	// Invoke the requested tool in the bottom panel
{
if ( ShowBottom() )							 												// If there's already a split, there's already a Tool instance
	{	tool = new Tools( this );															// If not, make one
		tool->Init( whichtool );																// whichtool informs Tools what it's supposed to be running eg grep, locate
	}
 else
 	{	if ( whichtool != terminalem )  tool->Init( whichtool );				// It's a change of tool, eg from grep to locate, or a do-it-again
			else tool->Close();																// Otherwise it's a second Ctrl-M, so toggle closed
	}
}

void LayoutWindows::CloseBottom()
{
delete tool; tool = NULL;
UnShowBottom();
}

//--------------------------------------------------------------------------------------------------------------------------------------

BEGIN_EVENT_TABLE(TabTemplateOverwriteDlg,wxDialog)	// This is for a dialog used in MyNotebook::SaveAsTemplate()
EVT_BUTTON(XRCID("Overwrite"),TabTemplateOverwriteDlg::OnOverwriteButton)
EVT_BUTTON(XRCID("SaveAs"),TabTemplateOverwriteDlg::OnSaveAsButton)
END_EVENT_TABLE()
//-----------------------------------------------------------------------------------------------------------------------

MyTab::MyTab( MyNotebook* dad, wxWindowID id/* = -1*/, int tabdatano /*=0*/, int TemplateNo /*=-1*/, const wxString& tabname/* = ""*/, const wxString& startdir0 /*=""*/, const wxString& startdir1 /*=""*/ )
						:	wxPanel(dad, id)
{
tabdata = new class Tabdata;												// Make a new tabdata
				// Load it with data for the requested page.  If this doesn't exist, sensible defaults will be substituted
tabdatanumber = tabdata->Load( tabdatano, TemplateNo );	// tabdatanumber is set to the ACTUAL tabdata loaded ie 0 if the requested one didn't exist

Create( startdir0, startdir1 );													// Do the actual construction
}
													// This is the Copy Ctor version; used by DuplicateTab()
MyTab::MyTab( Tabdata* tabdatatocopy, MyNotebook* dad, wxWindowID id/* = -1*/, int TemplateNo /*=-1*/, const wxString& tabname/* = ""*/ )
						:	wxPanel( dad, id, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL, wxT(""))
{
tabdata = new class Tabdata( *tabdatatocopy );			// Make a new tabdata, copying the passed version

Create();																			// Do the actual construction
}

void MyTab::Create( const wxString& startdir0 /*=""*/, const wxString& startdir1 /*=""*/ )
{
wxBoxSizer* panelsizer = new wxBoxSizer( wxVERTICAL );

m_splitterQuad = new wxSplitterWindow(this, -1, wxDefaultPosition, wxDefaultSize,  0);	// Splitter that holds quad panes splitters
m_splitterQuad->SetMinimumPaneSize( 20 );

panelsizer->Add(m_splitterQuad, 1, wxEXPAND);
SetSizer( panelsizer );
																		// Now create the  other splitters
m_splitterLeftTop = new DirSplitterWindow( m_splitterQuad, false );								// One for the left or top (flagged by false)
m_splitterRightBottom = new DirSplitterWindow( m_splitterQuad, true );							// & one for the right or bottom
m_splitterLeftTop->SetMinimumPaneSize( 20 );
m_splitterRightBottom->SetMinimumPaneSize( 20 );

																		// and do the splitting
#if wxVERSION_NUMBER > 2601 && ! defined(__WXX11__)  // For some reason, this isn't necessary/makes things worse in wxX11
	// Since 2.6.2 the sash positions are set when the window is still of zero size, so when it finally expands the Rt pane gets most of the area
PerformSplit( initialsplit, true );						// So do an initial split, then redo the user-defined dimensions from an idle event later
Connect( wxEVT_IDLE, (wxObjectEventFunction)&MyTab::OnIdle );
	#if wxVERSION_NUMBER > 2603 && ! (defined(__WXGTK20__) || defined(__WXX11__))
	NeedsLayout = false;
	#endif
#else
PerformSplit( tabdata->splittype, true );
#endif

wxString startdir_0 = startdir0, startdir_1 = startdir1;
if ( startdir_0.IsEmpty() ) startdir_0 = tabdata->startdirleft;		// If the user passed a "Start here" parameter or 2, override the tab's previous startdir
if ( startdir_1.IsEmpty() ) startdir_1 = tabdata->startdirright;
																		// Do the actual directory creation. True means starts with focus
m_splitterLeftTop->Setup( this, true, startdir_0, tabdata->fulltreeleft,
																			 tabdata->TLFileonly, tabdata->showhiddenTLdir, tabdata->showhiddenTLfile, tabdata->pathleft );
m_splitterRightBottom->Setup( this, false, startdir_1, tabdata->fulltreeright,
																			 tabdata->BRFileonly, tabdata->showhiddenBRdir, tabdata->showhiddenBRfile, tabdata->pathright );

#if wxVERSION_NUMBER > 2601
m_splitterQuad->SetSashGravity(0.5);			// Gravity was invented in 2602
m_splitterLeftTop->SetSashGravity(0.3);
m_splitterRightBottom->SetSashGravity(0.3);
#endif
}

void MyTab::PerformSplit( split_type splittype, bool newtab /*=false*/ )		// Splits a tab into left/right or top/bottom or unsplit, as desired
{
bool topleft = (LeftRight == LeftorTopDirCtrl);					// Which pane is currently active, Left(Top) or Right(Bottom)?  If unsplitting, we'll delete the other one.

int mode = m_splitterQuad->IsSplit();									// Are we currently split
if ( mode ) { mode = m_splitterQuad->GetSplitMode(); }	// If so which way?  (We have to do it like this as an unsplit SplitterWin reports it's already vertical)

switch( splittype )
	{	case vertical:			if ( mode == wxSPLIT_VERTICAL ) return;					// Nothing to do

										if ( mode == wxSPLIT_HORIZONTAL )							// Changing from horiz to vertical
											{	tabdata->horizontalsplitpos = m_splitterQuad->GetSashPosition();	// Save current sash positions, in case they'd been altered
												tabdata->tophorizontalsashpos = m_splitterLeftTop->GetSashPosition();
												tabdata->bottomhorizontalsashpos = m_splitterRightBottom->GetSashPosition();	
												m_splitterQuad->Unsplit( m_splitterRightBottom );	// Unsplit first.
												m_splitterQuad->SplitVertically(m_splitterLeftTop, m_splitterRightBottom, tabdata->verticalsplitpos);  
												m_splitterRightBottom->Show(true);						// Unsplitting hid this pane, so need to Show it again
																					// Set the desired sash positions, first checking there are valid nos
												if ( tabdata->leftverticalsashpos == -1 )					// If there is no valid sash position, set it to 1/3rd of the window
													{	wxSize size = m_splitterLeftTop->GetClientSize();
														tabdata->leftverticalsashpos = size.GetWidth() / 3;
													}
												if ( tabdata->rightverticalsashpos == -1 )
													{	wxSize size = m_splitterRightBottom->GetClientSize();
														tabdata->rightverticalsashpos = size.GetWidth() / 3;
													}

												m_splitterLeftTop->SetSashPosition( tabdata->leftverticalsashpos );
												m_splitterRightBottom->SetSashPosition( tabdata->rightverticalsashpos );
												splitstatus = vertical; return;
											}
																					// Otherwise we must be changing from unsplit to vertical, either by choice or because the tab is being created
										if ( tabdata->verticalsplitpos == 0 && newtab )					// First get a sensible default position if needed ie if new tab and no positive value
											{	wxSize size = MyFrame::mainframe->GetClientSize();	// (if not a newtab, a zero default halves the screen, which is OK)
												tabdata->verticalsplitpos = size.GetWidth();				// I'd've thought this should be halved, but this works best
											}

										m_splitterQuad->SplitVertically(m_splitterLeftTop, m_splitterRightBottom, tabdata->verticalsplitpos);  // Split the page into 2, vertically
										m_splitterLeftTop->Show(true); m_splitterRightBottom->Show(true); splitstatus = vertical;
										m_splitterLeftTop->SetSashPosition( tabdata->leftverticalsashpos );	// Make sure the sash positions haven't changed
										m_splitterRightBottom->SetSashPosition( tabdata->rightverticalsashpos );
										if ( ! newtab ) return;															// If this ISN'T a new tab, we're done	
																																	// Otherwise do the dirview/fileview splits
										if ( tabdata->leftverticalsashpos == -1 )								// Again check for valid defaults
											{	wxSize size = MyFrame::mainframe->GetClientSize();	// This time, use main window size, as the local clientsize isn't yet established
												tabdata->leftverticalsashpos = size.GetWidth() / 3;
											}
										if ( tabdata->rightverticalsashpos == -1 )
											{	wxSize size = MyFrame::mainframe->GetClientSize();
												tabdata->rightverticalsashpos = size.GetWidth() / 3;
											}
										m_splitterLeftTop->SplitVertically(m_splitterLeftTop->DirPanel, m_splitterLeftTop->FilePanel, tabdata->leftverticalsashpos );
										m_splitterRightBottom->SplitVertically(m_splitterRightBottom->DirPanel, m_splitterRightBottom->FilePanel, tabdata->rightverticalsashpos );
																								
										splitstatus = vertical; return;

										
		case horizontal:		if ( mode == wxSPLIT_HORIZONTAL )	return;
		
										if ( mode == wxSPLIT_VERTICAL)									// Changing from vertical to horiz.
											{	tabdata->verticalsplitpos = m_splitterQuad->GetSashPosition();	// Save current sash positions, in case they'd been altered
												tabdata->leftverticalsashpos = m_splitterLeftTop->GetSashPosition();
												tabdata->rightverticalsashpos = m_splitterRightBottom->GetSashPosition();

												m_splitterQuad->Unsplit( m_splitterRightBottom );	// Unsplit first
												m_splitterQuad->SplitHorizontally(m_splitterLeftTop, m_splitterRightBottom, tabdata->horizontalsplitpos);
												m_splitterRightBottom->Show(true);						// Unsplitting hid this pane, so need to Show it again
																	// Set the desired sash positions, first checking there are valid nos
												if ( tabdata->tophorizontalsashpos == -1 )				// If there is no valid sash position, set it to 1/3rd of the window
													{	wxSize size = m_splitterLeftTop->GetClientSize();
														tabdata->tophorizontalsashpos = size.GetWidth() / 3;
													}
												if ( tabdata->bottomhorizontalsashpos == -1 )
													{	wxSize size = m_splitterRightBottom->GetClientSize();
														tabdata->bottomhorizontalsashpos = size.GetWidth() / 3;
													}

												m_splitterLeftTop->SetSashPosition( tabdata->tophorizontalsashpos );
												m_splitterRightBottom->SetSashPosition( tabdata->bottomhorizontalsashpos );
												splitstatus = horizontal; return;
											}
												// Otherwise we must be changing from unsplit to horiz, either by choice or because the tab is being created
										if ( tabdata->horizontalsplitpos == 0  && newtab)			// First get a sensible default position if needed.  See 'vertical' comments
											{	wxSize size = MyFrame::mainframe->GetClientSize();
												tabdata->horizontalsplitpos = size.GetHeight();
											}

										m_splitterQuad->SplitHorizontally(m_splitterLeftTop, m_splitterRightBottom, tabdata->horizontalsplitpos);  // Split the page into 2, horiz.ly
										m_splitterRightBottom->Show(true); m_splitterLeftTop->Show(true); splitstatus = horizontal; 
										m_splitterLeftTop->SetSashPosition( tabdata->tophorizontalsashpos );
										m_splitterRightBottom->SetSashPosition( tabdata->bottomhorizontalsashpos );
										if ( ! newtab ) return;														// If this ISN'T a new tab, we're done	
																																// Otherwise do the dirview/fileview splits																	
										if ( tabdata->tophorizontalsashpos == -1 )
											{	wxSize size = MyFrame::mainframe->GetClientSize();
												tabdata->tophorizontalsashpos = size.GetWidth() / 3;
											}
										if ( tabdata->bottomhorizontalsashpos == -1 )
											{	wxSize size = MyFrame::mainframe->GetClientSize();
												tabdata->bottomhorizontalsashpos = size.GetWidth() / 3;
											}
										m_splitterLeftTop->SplitVertically(m_splitterLeftTop->DirPanel, m_splitterLeftTop->FilePanel, tabdata->tophorizontalsashpos );																																																				
										m_splitterRightBottom->SplitVertically(m_splitterRightBottom->DirPanel,
																																	m_splitterRightBottom->FilePanel, tabdata->bottomhorizontalsashpos );
										splitstatus = horizontal; return;

		case unsplit:			if ( mode == wxSPLIT_VERTICAL									// Changing from vertical to Unsplit
																|| mode == wxSPLIT_HORIZONTAL		//   or horizontal to Unsplit
																			|| splitstatus == initialsplit )			//     or previously just initialised
											{	if ( mode == wxSPLIT_VERTICAL )						// If vertical, save the vertical sash positions in case they'd changed
														{	tabdata->verticalsplitpos = m_splitterQuad->GetSashPosition();
															tabdata->leftverticalsashpos = m_splitterLeftTop->GetSashPosition();
															tabdata->rightverticalsashpos = m_splitterRightBottom->GetSashPosition();
														}
													else	{	if ( mode == wxSPLIT_HORIZONTAL )	// Similarly if horizontal
																	{	tabdata->horizontalsplitpos = m_splitterQuad->GetSashPosition();
																		tabdata->tophorizontalsashpos = m_splitterLeftTop->GetSashPosition();
																		tabdata->bottomhorizontalsashpos = m_splitterRightBottom->GetSashPosition();													
																	}
															}
												if ( topleft )	
													{	m_splitterQuad->Unsplit( m_splitterRightBottom );	// Unsplit the unwanted half
														m_splitterLeftTop->SetSashPosition( tabdata->unsplitsashpos );	// Set the sash to unsplit position
													}
												 else			
													{	m_splitterQuad->Unsplit( m_splitterLeftTop );
														m_splitterRightBottom->SetSashPosition( tabdata->unsplitsashpos );
													}
												splitstatus = unsplit; return;
											}

										if ( ! newtab ) return;														// If this ISN'T a new tab, nothing to do as it was already unsplit
										splitstatus = unsplit;														// Otherwise set the about-to-be status and fall thru into initialsplit
																										
		case initialsplit:		if ( tabdata->unsplitsashpos == -1 )								// Otherwise do the dirview/fileview splits
											{	wxSize size = MyFrame::mainframe->GetClientSize();
												tabdata->unsplitsashpos = size.GetWidth() / 2;
											}
										m_splitterQuad->Initialize( m_splitterLeftTop );
										m_splitterLeftTop->Show(true); m_splitterRightBottom->Show(true);
										
										m_splitterLeftTop->SplitVertically(m_splitterLeftTop->DirPanel, m_splitterLeftTop->FilePanel, tabdata->unsplitsashpos );
																		// We have to split m_splitterRightBottom here, even though it's invisible: it might be wanted later
										m_splitterRightBottom->SplitVertically(m_splitterRightBottom->DirPanel,
																																		m_splitterRightBottom->FilePanel, tabdata->unsplitsashpos );
										
										if ( splittype == initialsplit ) splitstatus = initialsplit;			// If we weren't falling thru from unsplit, set status.  It'll be needed when the idle event fires
										
	}
}

void MyTab::OnIdle( wxIdleEvent& WXUNUSED(event) )
{	// See the comment in MyTab::Create about delayed application of user-defined sizes in >2.6.1
#if wxVERSION_NUMBER > 2603 && ! (defined(__WXGTK20__) || defined(__WXX11__))
if ( NeedsLayout )						 // Recent versions don't do the initial layout properly in gtk1.2: only a few files are visible
	{	GetSizer()->Layout(); Disconnect( wxID_ANY, wxEVT_IDLE ); return; }				// If this is the second idle event, Layout, then Disconnect
#endif

if ( m_splitterQuad->GetClientSize().GetWidth() > 20 && splitstatus == initialsplit ) 		// Don't do it until the window system is up and running; but only once
	{	PerformSplit( tabdata->splittype );																			// OK, we're safe to do it properly
#if wxVERSION_NUMBER <= 2603 || defined(__WXGTK20__) || defined(__WXX11__)
		Disconnect( wxID_ANY, wxEVT_IDLE );																	// If this *isn't* a recent version/gtk1, Disconnect now: we don't want to keep being called
#else
		NeedsLayout = true;																								// Otherwise flag that we need to do Layout in the next idle event
#endif
	}
}

void MyTab::StoreData( int tabno )		// Refreshes the tabdata ready for saving
{
if ( tabdata == NULL )  return;

tabdata->startdirleft = LeftorTopDirCtrl->startdir;													// First store the various data in tabdata
tabdata->startdirright = RightorBottomDirCtrl->startdir;
tabdata->pathleft = LeftorTopDirCtrl->GetPath();
tabdata->pathright = RightorBottomDirCtrl->GetPath();
tabdata->fulltreeleft = LeftorTopDirCtrl->fulltree;
tabdata->fulltreeright = RightorBottomDirCtrl->fulltree;
//tabdata->tabname = MyFrame::mainframe->Layout->m_notebook->GetPageText( tabno );	// No, don't do this, otherwise we store Page1, Page2 etc. which should be auto-generated on loading
tabdata->showhiddenTLdir = LeftorTopDirCtrl->GetShowHidden();
tabdata->showhiddenTLfile = LeftorTopFileCtrl->GetShowHidden();
tabdata->showhiddenBRdir = RightorBottomDirCtrl->GetShowHidden();
tabdata->showhiddenBRfile = RightorBottomFileCtrl->GetShowHidden();

tabdata->TLFileonly = LeftorTopFileCtrl->DisplayFilesOnly;
tabdata->BRFileonly = RightorBottomFileCtrl->DisplayFilesOnly;

for ( size_t n=0; n < TOTAL_NO_OF_FILEVIEWCOLUMNS; ++n )
	{	tabdata->LColShown[n] = ! ((FileGenericDirCtrl*)LeftorTopFileCtrl)->GetHeaderWindow()->IsHidden(n);
		if ( tabdata->LColShown[n] )	 tabdata->Lwidthcol[n] = ((FileGenericDirCtrl*)LeftorTopFileCtrl)->GetHeaderWindow()->GetColumnWidth( n );
		  else  tabdata->Lwidthcol[n] = ((FileGenericDirCtrl*)LeftorTopFileCtrl)->GetHeaderWindow()->GetColumnDesiredWidth( n );	// If not currently shown, store the size it would have been

		tabdata->RColShown[n] = ! ((FileGenericDirCtrl*)RightorBottomFileCtrl)->GetHeaderWindow()->IsHidden(n);
		if ( tabdata->RColShown[n] )	 tabdata->Rwidthcol[n] = ((FileGenericDirCtrl*)RightorBottomFileCtrl)->GetHeaderWindow()->GetColumnWidth( n );
		  else  tabdata->Rwidthcol[n] = ((FileGenericDirCtrl*)RightorBottomFileCtrl)->GetHeaderWindow()->GetColumnDesiredWidth( n );
	}

tabdata->Lselectedcol = ((FileGenericDirCtrl*)LeftorTopFileCtrl)->GetHeaderWindow()->GetSelectedColumn();
tabdata->Rselectedcol = ((FileGenericDirCtrl*)RightorBottomFileCtrl)->GetHeaderWindow()->GetSelectedColumn();
tabdata->Lreversesort = ((FileGenericDirCtrl*)LeftorTopFileCtrl)->GetHeaderWindow()->GetSortOrder();
tabdata->Rreversesort = ((FileGenericDirCtrl*)RightorBottomFileCtrl)->GetHeaderWindow()->GetSortOrder();
	
int mode = m_splitterQuad->IsSplit();									// Are we currently split
if ( mode ) {mode = m_splitterQuad->GetSplitMode();	}	// If so which way?  (We have to do it like this as an unsplit SplitterWin reports it's already vertical)
switch( mode )
	{	case wxSPLIT_VERTICAL:			tabdata->splittype = vertical;
															tabdata->verticalsplitpos = m_splitterQuad->GetSashPosition();
															tabdata->leftverticalsashpos = m_splitterLeftTop->GetSashPosition();
															tabdata->rightverticalsashpos = m_splitterRightBottom->GetSashPosition();
															break;
		case wxSPLIT_HORIZONTAL:	tabdata->splittype = horizontal; 
															tabdata->horizontalsplitpos = m_splitterQuad->GetSashPosition();
															tabdata->tophorizontalsashpos = m_splitterLeftTop->GetSashPosition();
															tabdata->bottomhorizontalsashpos = m_splitterRightBottom->GetSashPosition();													
															break;
		 	default:									tabdata->splittype = unsplit;
															tabdata->unsplitsashpos = ((wxSplitterWindow*)m_splitterQuad->GetWindow1())->GetSashPosition();
	}
}

//-----------------------------------------------------------------------------------------------------------------------
Tabdata::Tabdata( const Tabdata& oldtabdata )
{
startdirleft = oldtabdata.startdirleft;									// Left(Top) pane starting dir and path
pathleft = oldtabdata.pathleft;
startdirright = oldtabdata.startdirright;								// Ditto Right(Bottom)
pathright = oldtabdata.pathright;

fulltreeleft = oldtabdata.fulltreeleft;									//  & their fulltree status
fulltreeright = oldtabdata.fulltreeright;

TLFileonly = oldtabdata.TLFileonly;								// & file-only status
BRFileonly = oldtabdata.BRFileonly;

showhiddenTLdir = oldtabdata.showhiddenTLdir;			// & showhidden status
showhiddenTLfile = oldtabdata.showhiddenTLfile;
showhiddenBRdir = oldtabdata.showhiddenBRdir;
showhiddenBRfile = oldtabdata.showhiddenBRfile;

splittype = oldtabdata.splittype; 										// How is the pane split
verticalsplitpos = oldtabdata.verticalsplitpos;						// If it's vertical, this is the main sash pos
horizontalsplitpos = oldtabdata.horizontalsplitpos;			// Horizontal ditto

leftverticalsashpos = oldtabdata.leftverticalsashpos;			// If it's vertical, this is where to position the dirview/fileview sash
tophorizontalsashpos = oldtabdata.tophorizontalsashpos;	// etc
rightverticalsashpos = oldtabdata.rightverticalsashpos;
bottomhorizontalsashpos = oldtabdata.bottomhorizontalsashpos;
unsplitsashpos = oldtabdata.unsplitsashpos;

for ( size_t n=0; n < TOTAL_NO_OF_FILEVIEWCOLUMNS; ++n )	// Widths of the 2 fileviews' columns
	{	Lwidthcol[n] = oldtabdata.Lwidthcol[n]; Rwidthcol[n] = oldtabdata.Rwidthcol[n];
		LColShown[n] = oldtabdata.LColShown[n]; RColShown[n] = oldtabdata.RColShown[n];
	}

Lselectedcol = oldtabdata.Lselectedcol;
Rselectedcol = oldtabdata.Rselectedcol;
Lreversesort = oldtabdata.Lreversesort;
Rreversesort = oldtabdata.Rreversesort;

tabname = oldtabdata.tabname;										// Label on the tab
}

int Tabdata::Load( int tabno /*= 0*/, int TemplateNo /*=-1*/ )	// Load the current defaults for this tab.  tabno allows different tabs in the notebook to have different defaults
{
int tabnumber = tabno; bool usingdefault = false;

wxConfigBase *config = wxConfigBase::Get();					// Find the config data
if (config==NULL)  { wxLogError(_("Couldn't load configuration!?")); return false; }

wxString Path;
if ( TemplateNo >= 0 ) Path.Printf( wxT("/Tabs/Template_%c/"), wxT('a') + TemplateNo );	// If we're loading a template, adjust path approprately
 else Path = wxT("/Tabs/");

wxString Subpath; Subpath.Printf(wxT("tab_%c/"), wxT('a') + tabno );			// Make the subpath read tab_a/, tab_b/ or whatever
if ( ! config->Exists( wxString(Path+Subpath).BeforeLast('/') ) )  
	{ Subpath = wxT("tab_a/"); tabnumber = 0; usingdefault = true; }			// If the requested tabno has no defaults, default to 0.  NB this won't work for templates with no tabs, but surely . . .

Path += Subpath;

wxString key, pageno; pageno.Printf(wxT("Page %u"),  tabno+1 );				// Make a default tab title
if ( usingdefault ) tabname = pageno;																// If we're using (probably) page 0 because there isn't a page 21, we mustn't load the page's title.
  else																												// If we did, we'd end up with a whole series tabs called "Page 1"
	{	key = wxT("tabname"); config->Read( Path+key, &tabname, pageno ); }

key = wxT("startdirleft"); config->Read( Path+key, &startdirleft, wxT("") );
key = wxT("pathleft"); config->Read( Path+key, &pathleft, wxT("") );
key = wxT("fulltreeleft"); config->Read( Path+key, &fulltreeleft, 0 );

key = wxT("startdirright"); config->Read( Path+key, &startdirright, wxT("") );
key = wxT("pathright"); config->Read( Path+key, &pathright, wxT("") );
key = wxT("fulltreeright"); config->Read( Path+key, &fulltreeright, 0 );

key = wxT("TLFileonly"); config->Read( Path+key, &TLFileonly, 0 );
key = wxT("BRFileonly"); config->Read( Path+key, &BRFileonly, 0 );

key = wxT("showhiddenTLdir"); config->Read( Path+key, &showhiddenTLdir, 1 );
key = wxT("showhiddenTLfile"); config->Read( Path+key, &showhiddenTLfile, 1 );
key = wxT("showhiddenBRdir"); config->Read( Path+key, &showhiddenBRdir, 1 );
key = wxT("showhiddenBRfile"); config->Read( Path+key, &showhiddenBRfile, 1 );

key = wxT("splittype"); splittype = (split_type)config->Read( Path+key, (long)vertical );
key = wxT("verticalsplitpos"); verticalsplitpos = (int)config->Read( Path+key, (long)450 );
key = wxT("horizontalsplitpos"); horizontalsplitpos = (int)config->Read( Path+key, (long)300 );
key = wxT("leftverticalsashpos"); leftverticalsashpos = (int)config->Read( Path+key, (long)-1 );	// -1 is a cue to measure the page-size
key = wxT("rightverticalsashpos"); rightverticalsashpos = (int)config->Read( Path+key, (long)-1 );
key = wxT("tophorizontalsashpos"); tophorizontalsashpos = (int)config->Read( Path+key, (long)-1 );
key = wxT("bottomhorizontalsashpos"); bottomhorizontalsashpos = (int)config->Read( Path+key, (long)-1 );
key = wxT("unsplitsashpos"); unsplitsashpos = (int)config->Read( Path+key, (long)-1 );

wxString Lkey = wxT("Lwidthcol"), Rkey = wxT("Rwidthcol");
wxString LShownkey = wxT("LColShown"), RShownkey = wxT("RColShown");
for ( size_t n=0; n < TOTAL_NO_OF_FILEVIEWCOLUMNS; ++n )
	{	wxString suffix = wxChar('0'+n);
		Lwidthcol[n] = (int)config->Read( Path+Lkey+suffix, 0l ); 
		Rwidthcol[n] = (int)config->Read( Path+Rkey+suffix, 0l );
		config->Read( Path+LShownkey+suffix, &LColShown[n], DEFAULT_COLUMN_TEMPLATE[n] ); 
		config->Read( Path+RShownkey+suffix, &RColShown[n], DEFAULT_COLUMN_TEMPLATE[n] );
	}
key = wxT("Lselectedcol"); Lselectedcol = (int)config->Read( Path+key, (long)0 );
key = wxT("Rselectedcol"); Rselectedcol = (int)config->Read( Path+key, (long)0 );
key = wxT("Lreversesort"); config->Read( Path+key, &Lreversesort, 0 );
key = wxT("Rreversesort"); config->Read( Path+key, &Rreversesort, 0 );

return tabnumber;																			// Return the actually-used tab number (may be zero if requested one doesn't exist)
}

void Tabdata::Save( int tabno /*= 0*/, int TemplateNo /*=-1*/ )	// Save the data for this tab.  If TemplateNo==-1, it's a normal save
{
wxConfigBase *config = wxConfigBase::Get();									// Find the config data
if (config==NULL)  { wxLogError(_("Couldn't load configuration!?")); return; }

wxString Path;
if ( TemplateNo >= 0 ) Path.Printf( wxT("/Tabs/Template_%c/"), wxT('a') + TemplateNo );	// If we're saving a template, adjust path approprately
 else Path = wxT("/Tabs/");
 
wxString Subpath; Subpath.Printf(wxT("tab_%c/"), wxT('a') + tabno );		// Make the subpath read tab_a/, tab_b/ or whatever
Path += Subpath;

config->Write( Path+wxT("tabname"), tabname );

config->Write( Path+wxT("startdirleft"), startdirleft );
config->Write( Path+wxT("pathleft"), pathleft );
config->Write( Path+wxT("fulltreeleft"), fulltreeleft );

config->Write( Path+wxT("startdirright"), startdirright );
config->Write( Path+wxT("pathright"), pathright );
config->Write( Path+wxT("fulltreeright"), fulltreeright );

config->Write( Path+wxT("TLFileonly"), TLFileonly );
config->Write( Path+wxT("BRFileonly"), BRFileonly );

config->Write( Path+wxT("showhiddenTLdir"), showhiddenTLdir );
config->Write( Path+wxT("showhiddenTLfile"), showhiddenTLfile );
config->Write( Path+wxT("showhiddenBRdir"), showhiddenBRdir );
config->Write( Path+wxT("showhiddenBRfile"), showhiddenBRfile );

config->Write( Path+wxT("splittype"), (long)splittype );
config->Write( Path+wxT("verticalsplitpos"), (long)verticalsplitpos );
config->Write( Path+wxT("horizontalsplitpos"), (long)horizontalsplitpos );
config->Write( Path+wxT("leftverticalsashpos"), (long)leftverticalsashpos );
config->Write( Path+wxT("rightverticalsashpos"), (long)rightverticalsashpos );
config->Write( Path+wxT("tophorizontalsashpos"), (long)tophorizontalsashpos );
config->Write( Path+wxT("bottomhorizontalsashpos"), (long)bottomhorizontalsashpos );
config->Write( Path+wxT("unsplitsashpos"), (long)unsplitsashpos );

wxString Lkey = wxT("Lwidthcol"), Rkey = wxT("Rwidthcol");
wxString LShownkey = wxT("LColShown"), RShownkey = wxT("RColShown");
for ( size_t n=0; n < TOTAL_NO_OF_FILEVIEWCOLUMNS; ++n )
	{	wxString suffix = wxChar('0'+n);
		config->Write( Path+Lkey+suffix, (long)Lwidthcol[n] );
		config->Write( Path+Rkey+suffix, (long)Rwidthcol[n] );
		config->Write( Path+LShownkey+suffix, LColShown[n] );
		config->Write( Path+RShownkey+suffix, RColShown[n] );
	}

config->Write( Path+wxT("Lselectedcol"), (long)Lselectedcol );	
config->Write( Path+wxT("Rselectedcol"), (long)Rselectedcol );	
config->Write( Path+wxT("Lreversesort"), Lreversesort );	
config->Write( Path+wxT("Rreversesort"), Rreversesort );	
}

//--------------------------------------------------------------------------------------------------------------------------------------
DirSplitterWindow::DirSplitterWindow( wxSplitterWindow* parent, bool right )
			 	: 	 wxSplitterWindow(parent, -1, wxDefaultPosition, wxDefaultSize,0),  isright(right)
{
m_parent = parent;

DirPanel = new MyPanel( this, NextID++, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL, wxT("DirPanel"));
FilePanel = new MyPanel( this, NextID++, wxDefaultPosition, wxDefaultSize, wxTAB_TRAVERSAL, wxT("FilePanel"));
	
DirSizer = new wxBoxSizer( wxVERTICAL );
FileSizer= new wxBoxSizer( wxVERTICAL );
DirToolbarSizer= new wxBoxSizer( wxVERTICAL );
}

void DirSplitterWindow::Setup( MyTab* parent_tab, bool hasfocus, wxString& startdir/*=null*/, bool fulltree/*=false*/, 
																	bool showfiles/*=false*/, bool showhiddendirs/*=true*/, bool showhiddenfiles/*=true*/, wxString& path /*=""*/ )
{
DirSizer->Add( DirToolbarSizer, 0, wxEXPAND );
																			// Create the twin controls
m_left = new DirGenericDirCtrl(DirPanel, NextID++, startdir, wxDefaultPosition, wxDefaultSize, wxSUNKEN_BORDER | wxDIRCTRL_SELECT_FIRST, fulltree);
m_right = new FileGenericDirCtrl(FilePanel, NextID++, m_left->startdir, wxDefaultPosition, wxDefaultSize, wxSUNKEN_BORDER | wxDIRCTRL_SELECT_FIRST);

m_left->ParentTab = parent_tab;												// Tell the ctrls which page owns them
m_right->ParentTab = parent_tab;
m_left->TakeYourPartners( m_right );											// Tell each ctrl where its partner lives
m_right->TakeYourPartners( m_left );
m_left->isright = isright;																// Tell each ctrl whether it's part of the left/top or right/bottom twin panes
m_right->isright = isright;
																			// Finish off the dirview
m_left->SetShowHidden( showhiddendirs );
m_left->FinishCreation();

	// Now do the fileview.  Note that we use m_left->startdir.  This is lest the original startdir didn't exist & a default was used, in which case the fileview would be blank
m_right->DisplayFilesOnly = showfiles;
m_right->SetShowHidden( showhiddenfiles );
m_right->FinishCreation();

m_left->CreateToolbar();															// Now ParentTab is valid, it's safe to create the dirview toolbar

if ( hasfocus ) 																				// If this is the LeftorTop pane-pair
	{	parent_tab->LeftorTopDirCtrl = m_left;								// Store the dirview address
		parent_tab->LeftorTopFileCtrl = m_right;							//  & fileview
		parent_tab->LeftRight = m_left;											// Tell tabpage that this is the default dirctrl (because it's the left hand one)
	}
 else
 	{	parent_tab->RightorBottomDirCtrl = m_left;
		parent_tab->RightorBottomFileCtrl = m_right;
	}
	
m_right->SetPath( path );															// Set path in the fileview
m_right->CreateColumns( hasfocus );										// Set up columns in the fileview.  hasfocus tells fileview whether to use tabdata's L or R col-widths

DirSizer->Add( m_left, 1, wxEXPAND );										// Put controls in the sizer
FileSizer->Add( m_right, 1, wxEXPAND );	

DirPanel->SetSizer( DirSizer ); FilePanel->SetSizer( FileSizer );
}

void DirSplitterWindow::OnSashPosChanging(wxSplitterEvent& event)
{
size_t currentsashpos = GetSashPosition();
int diff = currentsashpos - event.GetSashPosition();					// Is the sash shrinking the dirview or the fileview, & by how much
if ( !diff ) return;																			// Abort if it hasn't moved

if ( diff > 0 )																					// If it's the dirview that's shrinking, impose an arbitrary min size of 10, so that it doesn't disappear
	{ if ( (currentsashpos - diff) < 30 )  event.Veto(); return; }
																									// Otherwise it's the fileview
int w, h; GetClientSize(&w, &h);													// Find the whole-splitterwin size (to subtract currentsashpos from)
if ( (w - event.GetSashPosition()) < 50 )   event.Veto();				// Need a larger minimum, to allow for scrollbars
}

void  DirSplitterWindow::OnFullTree(wxCommandEvent& event){ m_left->OnFullTree(event); }			// Various toolbar methods, all passed to dirview
void  DirSplitterWindow::OnDirUp(wxCommandEvent& event){ m_left->OnDirUp(event); }
void  DirSplitterWindow::OnDirBack(wxCommandEvent& event){ m_left->OnDirBack(event); }
void  DirSplitterWindow::OnDirForward(wxCommandEvent& event){ m_left->OnDirForward(event); }
void  DirSplitterWindow::OnDirDropdown(wxCommandEvent& event){ m_left->OnDirDropdown(event); }
void  DirSplitterWindow::OnDirCtrlTBButton(wxCommandEvent& event){ m_left->OnDirCtrlTBButton(event); }

void DirSplitterWindow::DoToolbarUI(wxUpdateUIEvent& event){ m_left->DoToolbarUI(event); }

BEGIN_EVENT_TABLE(DirSplitterWindow,wxSplitterWindow)
	EVT_MENU(IDM_TOOLBAR_fulltree, DirSplitterWindow::OnFullTree)
	EVT_MENU(IDM_TOOLBAR_up, DirSplitterWindow::OnDirUp)
	EVT_BUTTON(XRCID("IDM_TOOLBAR_back"), DirSplitterWindow::OnDirBack)
	EVT_BUTTON(XRCID("IDM_TOOLBAR_smalldropdown1"),DirSplitterWindow::OnDirDropdown)
	EVT_BUTTON(XRCID("IDM_TOOLBAR_forward"), DirSplitterWindow::OnDirForward)
	EVT_BUTTON(XRCID("IDM_TOOLBAR_smalldropdown2"),DirSplitterWindow::OnDirDropdown)	
	EVT_MENU_RANGE(IDM_TOOLBAR_bmfirst, IDM_TOOLBAR_bmlast, DirSplitterWindow::OnDirCtrlTBButton)

	EVT_UPDATE_UI_RANGE(IDM_TOOLBAR_fulltree,IDM_TOOLBAR_up, DirSplitterWindow::DoToolbarUI)
	EVT_UPDATE_UI_RANGE(XRCID("IDM_TOOLBAR_back"),XRCID("IDM_TOOLBAR_forward"), DirSplitterWindow::DoToolbarUI)

	EVT_SPLITTER_SASH_POS_CHANGING(-1, DirSplitterWindow::OnSashPosChanging)
	EVT_SPLITTER_DCLICK(-1, DirSplitterWindow::OnDClick)
END_EVENT_TABLE()

